from __future__ import annotations

import asyncio
import contextlib
import functools
import importlib.metadata
import importlib.util
import shlex
import subprocess
import sys
import typing as t
from collections import defaultdict
from pathlib import Path
from tempfile import NamedTemporaryFile
from textwrap import dedent
from typing import ParamSpec

import dotenv
import rtoml
import tabulate
import typer
from pogo_core import squash
from pogo_core.migration import Migration, find_heads, read_sql_migration, topological_sort
from pogo_core.util import migrate, sql

from pogo_migrate import exceptions, yoyo
from pogo_migrate.config import Config, load_config
from pogo_migrate.context import Context
from pogo_migrate.util import get_editor, make_file

tempfile_prefix = "_tmp_pogonew"


def load_dotenv() -> None:
    dotenv.load_dotenv(
        dotenv.find_dotenv(usecwd=True),
        override=False,
    )


def _version_callback(*, value: bool) -> None:
    """Get current cli version."""
    if value:  # pragma: no cover
        version = importlib.metadata.version("pogo-migrate")
        typer.echo(f"pogo-migrate {version}")
        raise typer.Exit


def _callback(  # pragma: no cover
    _version: bool | None = typer.Option(  # noqa: FBT001
        None,
        "-v",
        "--version",
        callback=_version_callback,
        help="Print version and exit.",
    ),
) -> None: ...


app: typer.Typer = typer.Typer(name="pogo", callback=_callback)


P = ParamSpec("P")
R = t.TypeVar("R")


def handle_exceptions(context: Context) -> t.Callable[t.Callable[P, t.Awaitable[R]], t.Callable[P, t.Awaitable[R]]]:
    def inner(f: t.Callable[P, t.Awaitable[R]]) -> t.Callable[P, t.Awaitable[R]]:
        """Decorator to handle exceptions from migrations."""

        @functools.wraps(f)
        async def wrapped(*args: P.args, **kwargs: P.kwargs) -> R:
            """Wrapped function.

            Args:
            ----
                *args: The function args.
                **kwargs: The function kwargs.

            Returns:
            -------
                The function result.

            Raises:
            ------
                PostgresConnectionError: If the connection to postgres fails.
                UnexpectedPostgresError: If an unexpected error occurs.

            """
            try:
                return await f(*args, **kwargs)
            except typer.Exit:
                raise
            except Exception as e:
                context.stacktrace()
                context.error(str(e))
                raise typer.Exit(code=1) from e

        return wrapped

    return inner


@app.command("init")
def init(
    migrations_location: str = typer.Option("./migrations", "-m", "--migrations-location"),
    database_config: str = typer.Option("{POGO_DATABASE}", "-d", "--database-env-key"),
    schema: str | None = typer.Option("public", help="Default schema name."),
    verbose: int = typer.Option(
        0,
        "-v",
        "--verbose",
        help="Verbose output. Use multiple times to increase level of verbosity.",
        count=True,
        max=3,
    ),
) -> None:
    """Initiate pogo configuration.

    Create a migrations folder, and inject pogo configuration into
    pyproject.toml.
    """
    context = Context(verbose)

    @handle_exceptions(context)  # type: ignore[reportCallIssue]
    async def init_() -> None:
        pyproject = Path("pyproject.toml")
        if not pyproject.exists():
            pyproject.touch()

        with pyproject.open() as f:  # noqa: ASYNC230
            data = rtoml.load(f)

        if "tool" in data and "pogo" in data["tool"]:
            context.error("pogo already configured.")
            context.warning("\n".join(["", "[tool.pogo]"] + [f'{k} = "{v}"' for k, v in data["tool"]["pogo"].items()]))
            raise typer.Exit(code=1)

        cwd = Path.cwd().absolute()
        p = Path(migrations_location).resolve().absolute()

        try:
            loc = p.relative_to(cwd)
        except ValueError as e:
            context.error("migrations_location is not a child of current location.")
            raise typer.Exit(code=1) from e

        data = {
            "tool": {
                "pogo": {
                    "migrations": f"./{loc}",
                    "database_config": database_config,
                    "schema": schema,
                },
            },
        }

        config = rtoml.dumps(data, pretty=True)
        context.error(config)
        if typer.confirm(f"Write configuration to {pyproject.absolute()}"):
            loc.mkdir(exist_ok=True, parents=True)
            with pyproject.open("a") as f:  # noqa: ASYNC230
                f.write("\n")
                f.write(config)

    asyncio.run(init_())


migration_template: str = dedent(
    '''\
    """
    {message}
    """

    __depends__ = [{depends}]


    async def apply(db):
        ...


    async def rollback(db):
        ...
    ''',
)

migration_sql_template: str = dedent(
    """\
    --{message}
    -- depends:{depends}

    -- migrate: apply

    -- migrate: rollback

    """,
)


def retry(context: Context) -> str:
    choice = ""
    while choice == "":
        choice = typer.prompt("Retry editing? [Ynqh]", default="y", show_default=False)
        if choice == "q":
            raise typer.Exit(code=0)
        if choice in "yn":
            return choice
        if choice == "h":
            context.error("""\
y: reopen the migration file in your editor
n: save the migration as-is, without re-editing
q: quit without saving the migration
h: show this help
""")
        choice = ""

    # Unreachable
    return ""  # pragma: no cover


def create_with_editor(config: Config, content: str, extension: str, context: Context) -> Path:
    editor = get_editor()
    tmpfile = NamedTemporaryFile(  # noqa: SIM115
        mode="w",
        encoding="UTF-8",
        dir=config.migrations,
        prefix=tempfile_prefix,
        suffix=extension,
        delete=False,
    )

    orig_sys_path = sys.path[::]
    try:
        with tmpfile as f:
            f.write(content)
        editor = [part.format(tmpfile.name) for part in shlex.split(editor)]
        if not any(tmpfile.name in part for part in editor):
            editor.append(tmpfile.name)

        mtime = Path(tmpfile.name).lstat().st_mtime
        sys.path.insert(0, str(config.migrations))
        while True:
            try:
                subprocess.call(editor)  # noqa: S603
            except OSError as e:
                context.error("Error: could not open editor!")
                raise typer.Exit(code=1) from e
            else:
                if Path(tmpfile.name).lstat().st_mtime == mtime:
                    context.error("Abort: no changes made")
                    raise typer.Exit(code=1)

            try:
                migration = Migration("temporary", Path(tmpfile.name), None)
                migration.load()
                message = migration.__doc__
                break
            except Exception:  # noqa: BLE001
                context.stacktrace()
                context.error("Error loading migration.")
                choice = retry(context)
                if choice == "n":
                    message = ""
                    break

        filename = make_file(config.migrations, message, extension)
        Path(tmpfile.name).rename(filename)
        return filename
    finally:
        sys.path = orig_sys_path
        with contextlib.suppress(OSError):
            Path(tmpfile.name).unlink()


@app.command("new")
def new(
    message_: str = typer.Option("", "-m", "--message", help="Message describing focus of the migration."),
    *,
    interactive: bool = typer.Option(True, help="Open migration for editing."),  # noqa: FBT003
    py_: bool = typer.Option(False, "--py", help="Generate a python migration."),  # noqa: FBT003
    verbose: int = typer.Option(
        0,
        "-v",
        "--verbose",
        help="Verbose output. Use multiple times to increase level of verbosity.",
        count=True,
        max=3,
    ),
) -> None:
    """Generate a new migration."""
    context = Context(verbose)

    @handle_exceptions(context)  # type: ignore[reportCallIssue]
    async def new_() -> None:
        load_dotenv()
        config = load_config()

        migrations = await sql.read_migrations(config.migrations, db=None, schema_name="public")
        heads = find_heads([m.load() for m in migrations])

        template = migration_sql_template if not py_ else migration_template
        depends = sorted([m.id for m in heads])
        depends = f" {' '.join(depends)}" if not py_ and depends else depends
        depends = f'''"{'", "'.join(depends)}"''' if py_ and depends else (depends or "")
        message = f" {message_}" if not py_ and message_ else message_
        content = template.format(message=message, depends=depends)
        extension = ".sql" if not py_ else ".py"

        if not interactive:
            fp = make_file(config.migrations, message, extension)
            with fp.open("w", encoding="UTF-8") as f:
                f.write(content)
            raise typer.Exit(code=0)

        p = create_with_editor(config, content, extension, context)
        context.error("Created file: %s", p.as_posix().replace(config.root_directory.as_posix(), "").lstrip("/"))

    asyncio.run(new_())


@app.command("history")
def history(
    database: str | None = typer.Option(None, "-d", "--database", help="Database connection string."),
    *,
    schema: str | None = typer.Option(None, help="Schema override."),
    unapplied: bool = typer.Option(False, help="Show only unapplied migrations."),  # noqa: FBT003
    simple: bool = typer.Option(False, help="Show raw data without tabulation."),  # noqa: FBT003
    verbose: int = typer.Option(
        0,
        "-v",
        "--verbose",
        help="Verbose output. Use multiple times to increase level of verbosity.",
        count=True,
        max=3,
    ),
) -> None:
    """List migration history.

    If database location is configured, also check applied status of each migration.
    """

    context = Context(verbose)

    @handle_exceptions(context)  # type: ignore[reportCallIssue]
    async def history_() -> None:
        load_dotenv()
        config = load_config()

        try:
            connection_string = database or config.database_dsn
        except exceptions.InvalidConfigurationError:
            connection_string = database
        schema_name = schema or config.schema
        db = await sql.get_connection(connection_string, schema_name=schema_name) if connection_string else None

        if db is None:
            context.warning("Database connection can not be established, migration status can not be determined.")

        migrations = await sql.read_migrations(config.migrations, db, schema_name=schema_name)
        migrations = topological_sort([m.load() for m in migrations])

        data = (
            (
                "A" if m.applied else "U",
                m.id,
                "sql" if m.is_sql else "py",
            )
            for m in migrations
            if not unapplied or (unapplied and not m.applied)
        )
        if not simple:
            context.error(tabulate.tabulate(data, headers=("STATUS", "ID", "FORMAT")))
        else:
            for d in data:
                context.error(" ".join(d))

    asyncio.run(history_())


@app.command("apply")
def apply(
    database: str | None = typer.Option(None, "-d", "--database", help="Database connection string."),
    *,
    schema: str | None = typer.Option(None, help="Schema override."),
    create_schema: bool = typer.Option(False, "--create-schema/ ", help="Create database schema."),  # noqa: FBT003
    verbose: int = typer.Option(
        0,
        "-v",
        "--verbose",
        help="Verbose output. Use multiple times to increase level of verbosity.",
        count=True,
        max=3,
    ),
) -> None:
    """Apply migrations."""
    context = Context(verbose)

    @handle_exceptions(context)  # type: ignore[reportCallIssue]
    async def apply_() -> None:
        load_dotenv()
        config = load_config()

        connection_string = database or config.database_dsn
        schema_name = schema or config.schema
        db = await sql.get_connection(connection_string, schema_name=schema_name, schema_create=create_schema)

        await migrate.apply(db, config.migrations, schema_name=schema_name, logger=context)

    asyncio.run(apply_())


@app.command("rollback")
def rollback(
    database: str | None = typer.Option(None, "-d", "--database", help="Database connection string."),
    count: int = typer.Option(
        1,
        "-c",
        "--count",
        help="Number of migrations to rollback",
    ),
    *,
    schema: str | None = typer.Option(None, help="Schema override."),
    drop_schema: bool = typer.Option(
        False,  # noqa: FBT003
        "--drop-schema/ ",
        help="Drop database schema if all migrations rolled back.",
    ),
    verbose: int = typer.Option(
        0,
        "-v",
        "--verbose",
        help="Verbose output. Use multiple times to increase level of verbosity.",
        count=True,
        max=3,
    ),
) -> None:
    """Rollback one or more migrations."""
    context = Context(verbose)

    @handle_exceptions(context)  # type: ignore[reportCallIssue]
    async def rollback_() -> None:
        load_dotenv()
        config = load_config()

        connection_string = database or config.database_dsn
        schema_name = schema or config.schema
        db = await sql.get_connection(connection_string, schema_name=schema_name)

        await migrate.rollback(
            db,
            config.migrations,
            logger=context,
            count=count if count > 0 else None,
            schema_name=schema_name,
        )

        if drop_schema:
            applied_migrations = await sql.get_applied_migrations(db, schema_name=schema_name) if db else set()

            if len(applied_migrations) > 0:
                context.error("migrations still exist, can't drop schema.")
                raise typer.Exit(code=1)

            await sql.drop_schema(db, schema_name=schema_name)

    asyncio.run(rollback_())


@app.command("remove")
def remove(
    migration_id: str = typer.Argument(show_default=False, help="Migration id to remove (message can be excluded)."),
    migrations_location: str | None = typer.Option(None, "-m", "--migrations-location"),
    *,
    backup: bool = typer.Option(False, "--backup/ ", help="Keep .bak copy of original files."),  # noqa: FBT003
    verbose: int = typer.Option(
        0,
        "-v",
        "--verbose",
        help="Verbose output. Use multiple times to increase level of verbosity.",
        count=True,
        max=3,
    ),
) -> None:
    """Remove a migration from the dependency chain."""
    context = Context(verbose)

    load_dotenv()
    config = load_config()
    migrations_location = migrations_location or str(config.migrations)

    migrations = [
        Migration(path.stem, path, set())
        for path in Path(migrations_location).iterdir()
        if path.suffix in {".py", ".sql"}
    ]
    migrations = topological_sort([m.load() for m in migrations])
    for i, migration in enumerate(migrations):
        if migration.id.startswith(migration_id):
            next_migration = None
            with contextlib.suppress(IndexError):
                next_migration = migrations[i + 1]

            squash.remove(migration, next_migration, logger=context, backup=backup)


@app.command("squash")
def squash_(  # noqa: C901, PLR0912, PLR0915, PLR0913
    migrations_location: str | None = typer.Option(None, "-m", "--migrations-location"),
    *,
    backup: bool = typer.Option(False, "--backup/ ", help="Keep .bak copy of original files."),  # noqa: FBT003
    source: bool = typer.Option(False, "--source/ ", help="Add comment for source migration to each statement."),  # noqa: FBT003
    prompt_update: bool = typer.Option(False, "--update-prompt/ ", help="Confirm before including UPDATE statements."),  # noqa: FBT003
    prompt_skip: bool = typer.Option(
        False,  # noqa: FBT003
        "--skip-prompt/ ",
        help="Confirm before skipping unsquashable files, allow removal instead.",
    ),
    verbose: int = typer.Option(
        0,
        "-v",
        "--verbose",
        help="Verbose output. Use multiple times to increase level of verbosity.",
        count=True,
        max=3,
    ),
) -> None:
    """Squash migrations [EXPERIMENTAL].

    Python migrations and non transaction based transactions are skipped by default.

    Statements in sql migrations are grouped by table, and applied in the order
    tables where discovered.

    Rollback statements follow the reverse logic, the last table discovered is
    grouped first.
    """
    context = Context(verbose)

    load_dotenv()
    config = load_config()
    migrations_location = migrations_location or str(config.migrations)

    migrations = [
        Migration(path.stem, path, set())
        for path in Path(migrations_location).iterdir()
        if path.suffix in {".py", ".sql"}
    ]
    migrations = topological_sort([m.load() for m in migrations])

    applies = defaultdict(list)
    rollbacks = defaultdict(list)
    replaced = {}
    squashed = []
    depends = None
    latest = None
    for idx, migration in enumerate(migrations):
        skip = migration.id in config.squash.exclude
        if not migration.is_sql or not migration.use_transaction:
            if prompt_skip:
                view = typer.confirm(f"View unsquashable migration {migration.id}", default=True)
                if view:
                    content = migration.path.read_text()
                    context.error(content)

                remove = typer.confirm(f"Remove unsquashable migration {migration.id}", default=False)
                if remove:
                    next_migration = None
                    with contextlib.suppress(IndexError):
                        next_migration = migrations[idx + 1]

                    squash.remove(migration, next_migration, logger=context, backup=backup)

                    latest = migration
                    continue
            skip = True

        if skip:
            if (applies or rollbacks) and latest:
                new = squash.write(applies, rollbacks, latest, depends, squashed)
                replaced[latest.id]["new"] = new
                if new is None:
                    # New returns None if no update would actually occur (single file squash)
                    # Unmark the migration as replaced
                    del replaced[latest.id]
                applies = defaultdict(list)
                rollbacks = defaultdict(list)
                squashed = []
            depends = migration.id
            continue

        context.warning("Squashing %s", migration.id)
        squashed.append(migration.id)
        latest = migration
        replaced[migration.id] = {"orig": migration}

        _, _, _, _, _, apply_statements, rollback_statements = read_sql_migration(migration.path)
        for i, apply in enumerate(apply_statements):
            try:
                parsed = squash.parse_sqlglot(apply, logger=context)
            except squash.ParseError as e:
                context.stacktrace()
                context.error("%s: %s", migration.id, str(e))
                context.warning(apply)
                raise typer.Exit(code=1) from e

            if source:
                parsed.statement = f"{parsed.statement} -- source: {migration.id}"

            if parsed.statement_type in ("CREATE", "ALTER", "DROP"):
                if parsed.identifier:
                    applies[parsed.identifier].append(parsed.statement)
                else:
                    context.error("Can not extract table from DDL statement in migration %s", migration.id)
                    context.warning(apply)
                    raise typer.Exit(code=1)

            else:
                keep = True
                if parsed.statement_type == "UPDATE" and prompt_update:
                    context.error("")
                    with contextlib.suppress(IndexError):
                        context.error("   %s", apply_statements[i - 1])
                    context.error(">> %s", apply)
                    with contextlib.suppress(IndexError):
                        context.error("   %s", apply_statements[i + 1])
                    context.error("")
                    keep = typer.confirm("Include update statement", default=True)

                if keep:
                    applies["__data"].append(parsed.statement)

        rollbacks_ = defaultdict(list)
        for rollback in reversed(rollback_statements):
            try:
                parsed = squash.parse_sqlglot(rollback, logger=context)
            except squash.ParseError as e:
                context.stacktrace()
                context.error("%s: %s", migration.id, str(e))
                context.warning(rollback)
                raise typer.Exit(code=1) from e
            if source:
                parsed.statement = f"{parsed.statement} -- source: {migration.id}"

            if parsed.statement_type in ("CREATE", "ALTER", "DROP"):
                if parsed.identifier:
                    rollbacks_[parsed.identifier].append(parsed.statement)
                else:
                    context.error("Can not extract table from DDL statement in migration %s", migration.id)
                    context.warning(rollback)
                    raise typer.Exit(code=1)
            else:
                context.debug(parsed.statement_type)
                rollbacks_["__data"].append(parsed.statement)

        for ident, statements in rollbacks_.items():
            rollbacks[ident].append(statements)

    if (applies or rollbacks) and latest:
        new = squash.write(applies, rollbacks, latest, depends, squashed)
        replaced[latest.id]["new"] = new
        if new is None:
            # New returns None if no update would actually occur (single file squash)
            # Unmark the migration as replaced
            del replaced[latest.id]

    for data in replaced.values():
        orig, new = data["orig"], data.get("new")
        if backup:
            orig.path.rename(f"{orig.path}.bak")
        else:
            orig.path.unlink()
        if new:
            new.path.rename(orig.path)


@app.command("clean")
def clean(
    migrations_location: str | None = typer.Option(None, "-m", "--migrations-location"),
    *,
    verbose: int = typer.Option(
        0,
        "-v",
        "--verbose",
        help="Verbose output. Use multiple times to increase level of verbosity.",
        count=True,
        max=3,
    ),
) -> None:
    """Clean the migration directory of .bak migrations from squash."""
    _context = Context(verbose)
    load_dotenv()
    config = load_config()
    migrations_location = migrations_location or str(config.migrations)

    for path in Path(migrations_location).iterdir():
        if path.suffix in {".bak"}:
            path.unlink()


@app.command("validate")
def validate(
    migrations_location: str | None = typer.Option(None, "-m", "--migrations-location"),
    *,
    verbose: int = typer.Option(
        0,
        "-v",
        "--verbose",
        help="Verbose output. Use multiple times to increase level of verbosity.",
        count=True,
        max=3,
    ),
) -> None:
    """Validate migrations [EXPERIMENTAL].

    Best effort pass through to make sure identifiers aren't keywords.
    """
    context = Context(verbose)
    load_dotenv()
    config = load_config()
    migrations_location = migrations_location or str(config.migrations)

    migrations = [
        Migration(path.stem, path, set())
        for path in Path(migrations_location).iterdir()
        if path.suffix in {".py", ".sql"}
    ]
    migrations = topological_sort([m.load() for m in migrations])

    for migration in migrations:
        if not migration.is_sql:
            from unittest import mock  # noqa: PLC0415

            mock_asyncpg = mock.Mock()
            mock_asyncpg.execute = mock.AsyncMock(return_value=None)
            mock_asyncpg.fetch = mock.AsyncMock(return_value=[mock.MagicMock()])
            mock_asyncpg.fetchrow = mock.AsyncMock(return_value=mock.MagicMock())
            mock_asyncpg.fetchval = mock.AsyncMock(return_value=mock.MagicMock())
            try:
                asyncio.run(migration.apply(mock_asyncpg))
            except Exception:  # noqa: BLE001
                context.stacktrace()
                context.warning("Can't validate python migration %s (apply), skipping...", migration.id)

            try:
                asyncio.run(migration.rollback(mock_asyncpg))
            except Exception:  # noqa: BLE001
                context.stacktrace()
                context.warning("Can't validate python migration %s (rollback), skipping...", migration.id)

            statements = [c[1].get("query") or c[0][0] for c in mock_asyncpg.execute.call_args_list]
            statements += [c[1].get("query") or c[0][0] for c in mock_asyncpg.fetch.call_args_list]
            statements += [c[1].get("query") or c[0][0] for c in mock_asyncpg.fetchrow.call_args_list]
            statements += [c[1].get("query") or c[0][0] for c in mock_asyncpg.fetchval.call_args_list]
        else:
            _, _, _, _, _, apply_statements, rollback_statements = read_sql_migration(migration.path)
            statements = apply_statements + rollback_statements

        for statement in statements:
            try:
                parsed = squash.parse_sqlglot(statement, logger=context)
            except squash.ParseError as e:
                context.stacktrace()
                context.error("%s: %s", migration.id, str(e))
                context.warning(statement)
                continue

            if parsed.statement_type in ("CREATE", "ALTER", "DROP") and parsed.identifier is None:
                context.error(
                    "Can not extract table from DDL statement in migration %s, check that table name is not a reserved word.",
                    migration.id,
                )
                context.warning(statement)


@app.command("mark")
def mark(
    migration_id: str | None = typer.Option(None, "-m", "--migration", help="Specific migration to mark."),
    database: str | None = typer.Option(None, "-d", "--database", help="Database connection string."),
    *,
    schema: str | None = typer.Option(None, help="Schema override."),
    interactive: bool = typer.Option(True, help="Confirm all changes."),  # noqa: FBT003
    verbose: int = typer.Option(
        0,
        "-v",
        "--verbose",
        help="Verbose output. Use multiple times to increase level of verbosity.",
        count=True,
        max=3,
    ),
) -> None:
    """Mark migrations as applied, without running.

    If a migration_id is provided other migrations will be ignored.
    """
    context = Context(verbose)

    @handle_exceptions(context)  # type: ignore[reportCallIssue]
    async def _mark() -> None:
        load_dotenv()
        config = load_config()

        connection_string = database or config.database_dsn
        schema_name = schema or config.schema
        db = await sql.get_connection(connection_string, schema_name=schema_name)

        migrations = await sql.read_migrations(config.migrations, db, schema_name=schema_name)
        migrations = topological_sort([m.load() for m in migrations if migration_id is None or m.id == migration_id])

        async with db.transaction():
            for migration in migrations:
                migration.load()
                if not migration.applied:
                    if interactive and not typer.confirm(f"Mark {migration.id} as applied?"):
                        break

                    await sql.migration_applied(db, migration.id, migration.hash, schema_name=schema_name)

    asyncio.run(_mark())


@app.command("unmark")
def unmark(
    migration_id: str | None = typer.Option(None, "-m", "--migration", help="Specific migration to mark."),
    database: str | None = typer.Option(None, "-d", "--database", help="Database connection string."),
    *,
    schema: str | None = typer.Option(None, help="Schema override."),
    verbose: int = typer.Option(
        0,
        "-v",
        "--verbose",
        help="Verbose output. Use multiple times to increase level of verbosity.",
        count=True,
        max=3,
    ),
) -> None:
    """Mark migrations as unapplied, without rolling back.

    If a migration_id is provided other migrations will be ignored.
    """
    context = Context(verbose)

    @handle_exceptions(context)  # type: ignore[reportCallIssue]
    async def _unmark() -> None:
        load_dotenv()
        config = load_config()

        connection_string = database or config.database_dsn
        schema_name = schema or config.schema
        db = await sql.get_connection(connection_string, schema_name=schema_name)

        migrations = await sql.read_migrations(config.migrations, db, schema_name=schema_name)
        migrations = reversed(
            topological_sort([m.load() for m in migrations if migration_id is None or m.id == migration_id]),
        )

        async with db.transaction():
            for migration in migrations:
                migration.load()
                if migration.applied:
                    if not typer.confirm(f"Unmark {migration.id} as applied?"):
                        break

                    await sql.migration_unapplied(db, migration.id, schema_name=schema_name)

    asyncio.run(_unmark())


@app.command("migrate-yoyo")
def migrate_yoyo(
    database: str | None = typer.Option(None, "-d", "--database", help="Database connection string."),
    *,
    skip_files: bool = typer.Option(False, help="Skip file migration, just copy yoyo history."),  # noqa: FBT003
    verbose: int = typer.Option(
        0,
        "-v",
        "--verbose",
        help="Verbose output. Use multiple times to increase level of verbosity.",
        count=True,
        max=3,
    ),
) -> None:
    """Migrate existing 'yoyo' migrations to 'pogo'."""
    context = Context(verbose)

    @handle_exceptions(context)  # type: ignore[reportCallIssue]
    async def _migrate() -> None:
        load_dotenv()
        config = load_config()
        if not skip_files:
            for path in sorted(config.migrations.iterdir()):
                if path.name.endswith(".rollback.sql"):
                    continue

                if path.suffix == ".sql":
                    content = yoyo.convert_sql_migration(path)

                    with path.open("w") as f:
                        f.write(content)
                    context.error(
                        "Converted '%s' successfully.",
                        path.as_posix().replace(config.root_directory.as_posix(), "").lstrip("/"),
                    )
                else:
                    context.error(
                        "Python files can not be migrated reliably, please manually update '%s'.",
                        path.as_posix().replace(config.root_directory.as_posix(), "").lstrip("/"),
                    )
        else:
            context.debug("skip-files set, ignoring existing migration files.")

        connection_string = database or config.database_dsn
        db = await sql.get_connection(connection_string)

        await yoyo.copy_yoyo_migration_history(context, db)

    asyncio.run(_migrate())



import os
import random
import re
import string
import unicodedata
from datetime import datetime, timezone
from pathlib import Path


def unidecode(s: str) -> str:
    """
    Return ``s`` with unicode diacritics removed.
    """
    combining = unicodedata.combining
    return "".join(c for c in unicodedata.normalize("NFD", s) if not combining(c))


def slugify(message: str) -> str:
    s = unidecode(message)
    s = re.sub(r"[^-a-z0-9]+", "-", s.lower())
    return re.sub(r"-{2,}", "-", s).strip("-")


def random_string() -> str:
    """Generate a random 5 digit string."""
    return "".join(random.choices(string.digits + string.ascii_lowercase, k=5))  # noqa: S311


def make_file(migrations_dir: Path, message: str, extension: str) -> Path:
    slug = f"-{slugify(message)}" if message else ""
    datestr = datetime.now(tz=timezone.utc).date().strftime("%Y%m%d")
    rand = random_string()

    current = max(
        [int(p.name[len(datestr) + 1 :].split("_")[0]) for p in Path(migrations_dir).glob(f"{datestr}_*")],
        default=0,
    )

    number = str(int(current) + 1).zfill(2)

    return migrations_dir / f"{datestr}_{number}_{rand}{slug}{extension}"


def get_editor() -> str:
    """
    Return the user's preferred visual editor
    """
    for key in ["VISUAL", "EDITOR"]:
        editor = os.environ.get(key, None)
        if editor:
            return editor
    return "vi"



import pytest

import pogo_migrate.config
from pogo_migrate import exceptions, migrate


@pytest.fixture
def pyproject(pyproject_factory):
    return pyproject_factory()


@pytest.fixture
def config(pyproject):  # noqa: ARG001
    return pogo_migrate.config.load_config()


@pytest.fixture
def _migration_one(migrations):
    p = migrations / "20240317_01_abcde-initial-migration.sql"

    with p.open("w") as f:
        f.write("""
-- initial migration
-- depends:

-- migrate: apply
CREATE TABLE table_one();

-- migrate: rollback
DROP TABLE table_one;
""")


@pytest.fixture
def _migration_two(migrations, _migration_one):
    p = migrations / "20240317_02_12345-second-migration.sql"

    with p.open("w") as f:
        f.write("""
-- second migration
-- depends: 20240317_01_abcde-initial-migration
-- transaction: false

-- migrate: apply
CREATE TABLE table_two();

-- Auto generated
-- From psql

-- migrate: rollback

-- Auto generated
-- From psql
DROP TABLE table_two;
""")


@pytest.fixture
def _broken_apply(migrations, _migration_two):
    p = migrations / "20240318_01_12345-broken-apply.sql"

    with p.open("w") as f:
        f.write("""
-- broker migration
-- depends: 20240317_02_12345-second-migration

-- migrate: apply
CREATE TABLE table_three;

-- migrate: rollback
DROP TABLE table_two;
""")


@pytest.fixture
def _broken_rollback(migrations, _migration_two):
    p = migrations / "20240318_01_12345-broken-rollback.sql"

    with p.open("w") as f:
        f.write("""
-- broker migration
-- depends: 20240317_02_12345-second-migration

-- migrate: apply
CREATE TABLE table_three();

-- migrate: rollback
DROP TABLE table_four;
""")


class Base:
    async def assert_tables(self, db_session, tables):
        stmt = """
        SELECT tablename
        FROM pg_tables
        WHERE  schemaname = 'public'
        ORDER BY tablename
        """
        results = await db_session.fetch(stmt)

        assert [r["tablename"] for r in results] == tables


class TestApply(Base):
    @pytest.mark.usefixtures("migrations")
    async def test_no_migrations_applies_pogo_tables(self, config, db_session, context):
        await migrate.apply(db_session, config.migrations, logger=context)

        await self.assert_tables(db_session, ["_pogo_migration", "_pogo_version"])

    @pytest.mark.usefixtures("_migration_two")
    async def test_migrations_applied(self, config, db_session, context):
        await migrate.apply(db_session, config.migrations, logger=context)

        await self.assert_tables(db_session, ["_pogo_migration", "_pogo_version", "table_one", "table_two"])

    @pytest.mark.usefixtures("_migration_two")
    async def test_already_applied_skips(self, config, db_session, context):
        await migrate.apply(db_session, config.migrations, logger=context)
        await migrate.apply(db_session, config.migrations, logger=context)

        await self.assert_tables(db_session, ["_pogo_migration", "_pogo_version", "table_one", "table_two"])

    @pytest.mark.usefixtures("_broken_apply")
    async def test_broken_migration_not_applied(self, config, db_session, context):
        with pytest.raises(exceptions.BadMigrationError) as e:
            await migrate.apply(db_session, config.migrations, logger=context)

        await self.assert_tables(db_session, ["_pogo_migration", "_pogo_version", "table_one", "table_two"])
        assert str(e.value) == "Failed to apply 20240318_01_12345-broken-apply"


class TestRollback(Base):
    @pytest.mark.usefixtures("migrations")
    async def test_no_migrations_applies_pogo_tables(self, config, db_session, context):
        await migrate.rollback(db_session, config.migrations, logger=context)

        await self.assert_tables(db_session, ["_pogo_migration", "_pogo_version"])

    @pytest.mark.usefixtures("_migration_two")
    async def test_latest_removed(self, config, db_session, context):
        await migrate.apply(db_session, config.migrations, logger=context)
        await migrate.rollback(db_session, config.migrations, logger=context, count=1)

        await self.assert_tables(db_session, ["_pogo_migration", "_pogo_version", "table_one"])

    @pytest.mark.usefixtures("_migration_two")
    async def test_all_removed(self, config, db_session, context):
        await migrate.apply(db_session, config.migrations, logger=context)
        await migrate.rollback(db_session, config.migrations, logger=context)

        await self.assert_tables(db_session, ["_pogo_migration", "_pogo_version"])

    @pytest.mark.usefixtures("_broken_rollback")
    async def test_broken_rollback_rollsback(self, config, db_session, context):
        await migrate.apply(db_session, config.migrations, logger=context)
        with pytest.raises(exceptions.BadMigrationError) as e:
            await migrate.rollback(db_session, config.migrations, logger=context, count=1)

        await self.assert_tables(
            db_session,
            ["_pogo_migration", "_pogo_version", "table_one", "table_three", "table_two"],
        )
        assert str(e.value) == "Failed to rollback 20240318_01_12345-broken-rollback"



from unittest import mock

import pytest
from pogo_core.util import sql as sql_core

from pogo_migrate import sql


async def assert_tables(db_session, tables):
    stmt = """
    SELECT tablename
    FROM pg_tables
    WHERE  schemaname = 'public'
    ORDER BY tablename
    """
    results = await db_session.fetch(stmt)

    assert [r["tablename"] for r in results] == tables


@pytest.mark.nosync
@pytest.mark.parametrize("schema", [None, "unit"])
async def test_ensure_pogo_sync_creates_tables(db_session, schema):
    if schema:
        await db_session.execute(f"CREATE SCHEMA {schema}")
        await db_session.execute(f"SET search_path = '{schema}'")

    await sql.ensure_pogo_sync(db_session)

    await assert_tables(db_session, ["_pogo_migration", "_pogo_version"])


@pytest.mark.nosync
@pytest.mark.parametrize("schema", [None, "unit"])
async def test_ensure_pogo_sync_handles_existing_tables(db_session, schema):
    if schema:
        await db_session.execute(f"CREATE SCHEMA {schema}")
        await db_session.execute(f"SET search_path = '{schema}'")

    await sql.ensure_pogo_sync(db_session)
    await sql.ensure_pogo_sync(db_session)

    await assert_tables(db_session, ["_pogo_migration", "_pogo_version"])


@pytest.mark.parametrize("schema", [None, "unit"])
async def test_migration_applied(db_session, schema):
    if schema:
        await db_session.execute(f"CREATE SCHEMA {schema}")
        await db_session.execute(f"SET search_path = '{schema}'")

    await sql.migration_applied(db_session, "migration_id", "migration_hash")

    ids = await sql.get_applied_migrations(db_session)
    assert ids == {"migration_id"}


@pytest.mark.parametrize("schema", [None, "unit"])
async def test_migration_unapplied(db_session, schema):
    if schema:
        await db_session.execute(f"CREATE SCHEMA {schema}")
        await db_session.execute(f"SET search_path = '{schema}'")

    await sql.migration_applied(db_session, "migration_id", "migration_hash")
    await sql.migration_unapplied(db_session, "migration_id")

    ids = await sql.get_applied_migrations(db_session)
    assert ids == set()


async def test_get_connection(db_session, monkeypatch):
    monkeypatch.setattr(sql_core.asyncpg, "connect", mock.AsyncMock(return_value=db_session))
    db = await sql.get_connection("connection-string")

    assert db == db_session


async def test_read_migrations(db_session, cwd):
    migrations = await sql.read_migrations(cwd, db_session)
    assert migrations == []



from __future__ import annotations

import typing as t
from warnings import warn

from pogo_core.util import sql

if t.TYPE_CHECKING:
    from pathlib import Path

    import asyncpg
    from pogo_core.migration import Migration


warn(
    "pogo_migrate.sql usage has been deprecated, please use pogo_core.util.sql",
    FutureWarning,
    stacklevel=2,
)


async def get_connection(
    connection_string: str,
    *,
    schema_name: str = "public",
    schema_create: bool = False,
) -> asyncpg.Connection:
    return await sql.get_connection(connection_string, schema_name=schema_name, schema_create=schema_create)


async def read_migrations(
    migrations_location: Path,
    db: asyncpg.Connection | None,
    *,
    schema_name: str = "public",
) -> list[Migration]:
    return await sql.read_migrations(migrations_location, db, schema_name=schema_name)


async def get_applied_migrations(db: asyncpg.Connection, *, schema_name: str = "public") -> set[str]:
    return await sql.get_applied_migrations(db, schema_name=schema_name)


async def ensure_pogo_sync(db: asyncpg.Connection) -> None:
    return await sql.ensure_pogo_sync(db)


async def migration_applied(
    db: asyncpg.Connection,
    migration_id: str,
    migration_hash: str,
    *,
    schema_name: str = "public",
) -> None:
    return await sql.migration_applied(db, migration_id, migration_hash, schema_name=schema_name)


async def migration_unapplied(db: asyncpg.Connection, migration_id: str, *, schema_name: str = "public") -> None:
    return await sql.migration_unapplied(db, migration_id, schema_name=schema_name)



from __future__ import annotations

import typing as t
from warnings import warn

from pogo_core import squash

if t.TYPE_CHECKING:
    from pogo_core.migration import Migration

    from pogo_migrate.context import Context


squash_sql_template: str = squash.squash_sql_template

warn(
    "pogo_migrate.squash usage has been deprecated, please use pogo_core.squash",
    FutureWarning,
    stacklevel=2,
)


def remove(context: Context, current: Migration, dependent: Migration | None, *, backup: bool = False) -> None:
    return squash.remove(current, dependent, context, backup=backup)


def write(
    apply_statements: dict[str, list[str]],
    rollback_statements: dict[str, list[list[str]]],
    latest: Migration,
    depends: str | None,
    squashed: list[str],
) -> Migration | None:
    return squash.write(apply_statements, rollback_statements, latest, depends, squashed)


ParsedStatement = squash.ParsedStatement


def parse(context: Context, statement: str) -> ParsedStatement:
    return squash.parse(statement, context)


ParseError = squash.ParseError


def parse_sqlglot(context: Context, statement: str) -> ParsedStatement:
    return squash.parse_sqlglot(statement, context)



from __future__ import annotations

import logging
import typing as t
from warnings import warn

from pogo_core.util import migrate

if t.TYPE_CHECKING:
    from pathlib import Path

    import asyncpg

    from pogo_migrate.context import Context

logger_ = logging.getLogger(__name__)

warn(
    "pogo_migrate.migrate usage has been deprecated, please use pogo_core.util.migrate",
    FutureWarning,
    stacklevel=2,
)


async def apply(
    db: asyncpg.Connection,
    migrations_dir: Path,
    *,
    schema_name: str = "public",
    logger: Context | logging.Logger | None = None,
) -> None:
    return await migrate.apply(db, migrations_dir, schema_name=schema_name, logger=logger)


async def rollback(
    db: asyncpg.Connection,
    migrations_dir: Path,
    *,
    schema_name: str = "public",
    count: int | None = None,
    logger: Context | logging.Logger | None = None,
) -> None:
    return await migrate.rollback(db, migrations_dir, schema_name=schema_name, count=count, logger=logger)



from textwrap import dedent

from pogo_migrate import yoyo


def test_convert_sql_migration_no_rollback(cwd):
    p = cwd / "20200101_01_slug-commit-message.sql"
    with p.open("w") as f:
        f.write(
            dedent("""\
        -- commit message
        -- depends: 20190101_01_slug2-initial-commit

        CREATE TABLE table_one();
        """),
        )

    content = yoyo.convert_sql_migration(p)
    assert content == dedent("""\
    -- commit message
    -- depends: 20190101_01_slug2-initial-commit

    -- migrate: apply

    CREATE TABLE table_one();

    -- migrate: rollback

    """)


def test_convert_sql_migration_with_rollback(cwd):
    p = cwd / "20200101_01_slug-commit-message.sql"
    with p.open("w") as f:
        f.write(
            dedent("""\
        CREATE TABLE table_one();
        """),
        )
    rp = cwd / "20200101_01_slug-commit-message.rollback.sql"
    with rp.open("w") as f:
        f.write(
            dedent("""\
        -- commit message
        -- depends: 20190101_01_slug2-initial-commit

        DROP TABLE table_one;
        """),
        )

    content = yoyo.convert_sql_migration(p)
    assert content == dedent("""\
    --
    -- depends:

    -- migrate: apply

    CREATE TABLE table_one();

    -- migrate: rollback

    DROP TABLE table_one;""")



"""Collection of useful commands for code management.

To view a list of available commands:

$ invoke --list
"""

import invoke  # type: ignore[unresolved-import]


@invoke.task
def install(context):
    """Install production requirements."""
    context.run("uv sync")


@invoke.task
def install_dev(context):
    """Install development requirements."""
    context.run("uv sync --all-extras")
    context.run("uv run pre-commit install")


@invoke.task
def check_style(context):
    """Run style checks."""
    context.run("ruff check .")


@invoke.task
def tests(context):
    """Run pytest unit tests."""
    context.run("pytest -x -s")


@invoke.task
def tests_coverage(context):
    """Run pytest unit tests with coverage."""
    context.run("pytest --cov -x --cov-report=xml")


@invoke.task
def infra_test_start(context):
    """Run local unittest infrastructure."""
    context.run(
        "docker compose -f unittest-compose.yml up -d",
    )


@invoke.task
def infra_test_stop(context):
    """Stop local unittest infrastructure."""
    context.run(
        "docker compose -f unittest-compose.yml down",
    )



from __future__ import annotations

import typing as t

from pogo_core.util import testing

from pogo_migrate import config

if t.TYPE_CHECKING:
    import asyncpg


async def apply(db: asyncpg.Connection | None = None) -> None:
    c = config.load_config()
    await testing.apply(c.migrations, db=db, database_dsn=c.database_dsn, schema_name=c.schema)


async def rollback(db: asyncpg.Connection | None = None) -> None:
    c = config.load_config()
    await testing.rollback(c.migrations, db=db, database_dsn=c.database_dsn, schema_name=c.schema)



import importlib.metadata
from pathlib import Path
from textwrap import dedent
from unittest import mock

import pytest
from pogo_core.util import sql

from pogo_migrate import cli
from tests.util import AsyncMock


def test_version(cli_runner):
    version = importlib.metadata.version("pogo-migrate")
    result = cli_runner.invoke(["--version"])
    assert result.exit_code == 0

    cli_runner.assert_output(f"pogo-migrate {version}")


@pytest.fixture
def pyproject(pyproject_factory, migrations):  # noqa: ARG001
    return pyproject_factory()


@pytest.fixture
def pyproject_no_database(pyproject_factory, migrations):  # noqa: ARG001
    return pyproject_factory(
        {
            "migrations": "./migrations",
        },
    )


@pytest.fixture(autouse=True)
def _db_patch(db_session, monkeypatch):
    monkeypatch.setattr(sql.asyncpg, "connect", AsyncMock(return_value=db_session))


class TestInit:
    def test_init_no_confirm(self, cwd, cli_runner):
        result = cli_runner.invoke(["init"], input="n\n")
        assert result.exit_code == 0, result.output

        p = cwd / "pyproject.toml"
        with p.open() as f:
            assert f.read() == ""

    def test_init_no_pyproject(self, cwd, cli_runner):
        result = cli_runner.invoke(["init"], input="y\n")
        assert result.exit_code == 0, result.output

        p = cwd / "pyproject.toml"
        with p.open() as f:
            assert f.read() == dedent("""\

            [tool.pogo]
            migrations = './migrations'
            database_config = '{POGO_DATABASE}'
            schema = 'public'
            """)

    def test_init_invalid_migrations_location(self, cwd, cli_runner):
        result = cli_runner.invoke(["init", "-m", str(cwd.parent / "migrations")])
        assert result.exit_code == 1, result.output
        cli_runner.assert_output(
            dedent("migrations_location is not a child of current location."),
        )

    def test_init_existing_pyproject(self, cwd, cli_runner):
        p = cwd / "pyproject.toml"
        with p.open("w") as f:
            f.write(
                dedent("""\
            [tool.other]
            key = "value"
            """),
            )
        result = cli_runner.invoke(["init"], input="y\n")
        assert result.exit_code == 0, result.output

        with p.open() as f:
            assert f.read() == dedent("""\
            [tool.other]
            key = "value"

            [tool.pogo]
            migrations = './migrations'
            database_config = '{POGO_DATABASE}'
            schema = 'public'
            """)

    def test_init_overrides(self, cwd, cli_runner):
        result = cli_runner.invoke(
            ["init", "-m", "./my-migrations", "-d", "{POSTGRES_DSN}", "--schema", "pogo"],
            input="y\n",
        )
        assert result.exit_code == 0, result.output

        p = cwd / "pyproject.toml"
        with p.open() as f:
            assert f.read() == dedent("""\

            [tool.pogo]
            migrations = './my-migrations'
            database_config = '{POSTGRES_DSN}'
            schema = 'pogo'
            """)

    def test_init_already_configured(self, cwd, cli_runner):
        p = cwd / "pyproject.toml"
        with p.open("w") as f:
            assert f.write(
                dedent("""\

            [tool.pogo]
            migrations = './my-migrations'
            database_config = '{POSTGRES_DSN}'
            """),
            )

        result = cli_runner.invoke(["init"])
        assert result.exit_code == 1
        cli_runner.assert_output(
            dedent("pogo already configured."),
        )

        p = cwd / "pyproject.toml"
        with p.open() as f:
            assert f.read() == dedent("""\

            [tool.pogo]
            migrations = './my-migrations'
            database_config = '{POSTGRES_DSN}'
            """)

    def test_init_already_configured_verbose(self, cwd, cli_runner):
        p = cwd / "pyproject.toml"
        with p.open("w") as f:
            assert f.write(
                dedent("""\

            [tool.pogo]
            migrations = './my-migrations'
            database_config = '{POSTGRES_DSN}'
            schema = 'public'
            """),
            )

        result = cli_runner.invoke(["init", "-v"])
        assert result.exit_code == 1
        cli_runner.assert_output(
            dedent("""\
            pogo already configured.

            [tool.pogo]
            migrations = "./my-migrations"
            database_config = "{POSTGRES_DSN}"
            schema = "public"
            """),
        )


class TestNew:
    def test_no_config(self, cli_runner):
        result = cli_runner.invoke(["new"])
        assert result.exit_code == 1, result.output
        cli_runner.assert_output(
            "No configuration found, missing pyproject.toml, run 'pogo init ...'",
        )

    @pytest.mark.usefixtures("pyproject")
    def test_non_interactive_file_written(self, monkeypatch, cli_runner, cwd):
        monkeypatch.setattr(cli, "make_file", mock.Mock(return_value=cwd / "new_file.py"))

        result = cli_runner.invoke(["new", "--no-interactive"])

        assert result.exit_code == 0, result.output
        with (cwd / "new_file.py").open() as f:
            assert f.read() == dedent("""\
            --
            -- depends:

            -- migrate: apply

            -- migrate: rollback

            """)

    @pytest.mark.usefixtures("pyproject_no_database")
    def test_non_interactive_file_written_optional_database(self, monkeypatch, cli_runner, cwd):
        monkeypatch.setattr(cli, "make_file", mock.Mock(return_value=cwd / "new_file.py"))

        result = cli_runner.invoke(["new", "--no-interactive"])

        assert result.exit_code == 0, result.output
        with (cwd / "new_file.py").open() as f:
            assert f.read() == dedent("""\
            --
            -- depends:

            -- migrate: apply

            -- migrate: rollback

            """)

    @pytest.mark.usefixtures("pyproject")
    def test_os_error(self, monkeypatch, cli_runner):
        monkeypatch.setattr(cli.subprocess, "call", mock.Mock(side_effect=OSError))
        result = cli_runner.invoke(["new"])
        assert result.exit_code == 1, result.output
        cli_runner.assert_output(
            "Error: could not open editor!",
        )

    @pytest.mark.usefixtures("pyproject")
    def test_no_changes(self, monkeypatch, cli_runner):
        monkeypatch.setattr(cli.subprocess, "call", mock.Mock())
        result = cli_runner.invoke(["new"])
        assert result.exit_code == 1, result.output
        cli_runner.assert_output(
            "Abort: no changes made",
        )

    @pytest.mark.usefixtures("pyproject")
    def test_changes_made(self, monkeypatch, cli_runner, migrations):
        monkeypatch.setattr(cli, "make_file", mock.Mock(return_value=migrations / "new_file.sql"))
        monkeypatch.setattr(cli.subprocess, "call", mock.Mock())
        monkeypatch.setattr(cli.Path, "lstat", mock.Mock(side_effect=[mock.Mock(), mock.Mock()]))

        result = cli_runner.invoke(["new", "-v"])

        assert result.exit_code == 0, result.output
        cli_runner.assert_output(
            dedent("""\
            Created file: migrations/new_file.sql
            """).strip(),
        )

    @pytest.mark.usefixtures("pyproject")
    def test_subprocess_call(self, monkeypatch, tmp_path, cli_runner):
        f = tmp_path / "tmpfile"
        f.touch()
        mock_file = mock.MagicMock()
        mock_file.name = str(f)

        monkeypatch.setenv("EDITOR", "vim")
        monkeypatch.setattr(cli.subprocess, "call", mock.Mock())
        monkeypatch.setattr(cli, "NamedTemporaryFile", mock.Mock(return_value=mock_file))

        cli_runner.invoke(["new", "--py"])

        assert cli.subprocess.call.call_args == mock.call(["vim", str(f)])

    @pytest.mark.usefixtures("pyproject")
    def test_subprocess_call_dynamic_editor(self, monkeypatch, tmp_path, cli_runner):
        f = tmp_path / "tmpfile"
        f.touch()
        mock_file = mock.MagicMock()
        mock_file.name = str(f)

        monkeypatch.setenv("EDITOR", "vim {}")
        monkeypatch.setattr(cli.subprocess, "call", mock.Mock())
        monkeypatch.setattr(cli, "NamedTemporaryFile", mock.Mock(return_value=mock_file))

        cli_runner.invoke(["new", "--py"])

        assert cli.subprocess.call.call_args == mock.call(["vim", str(f)])

    @pytest.mark.usefixtures("pyproject")
    @pytest.mark.parametrize(
        ("extra_args", "expected"),
        [
            (
                [],
                dedent("""\
                --
                -- depends:

                -- migrate: apply

                -- migrate: rollback

                """),
            ),
            (
                ["--py"],
                dedent('''\
                """

                """

                __depends__ = []


                async def apply(db):
                    ...


                async def rollback(db):
                    ...
                '''),
            ),
        ],
    )
    def test_file_written(self, monkeypatch, cli_runner, cwd, extra_args, expected):
        monkeypatch.setattr(cli, "make_file", mock.Mock(return_value=cwd / "new_file"))
        monkeypatch.setattr(cli.subprocess, "call", mock.Mock())
        monkeypatch.setattr(cli.Path, "lstat", mock.Mock(side_effect=[mock.Mock(), mock.Mock()]))

        result = cli_runner.invoke(["new", *extra_args])

        assert result.exit_code == 0, result.output
        with (cwd / "new_file").open() as f:
            assert f.read() == expected

    @pytest.mark.usefixtures("pyproject")
    @pytest.mark.parametrize(
        ("extra_args", "expected"),
        [
            (
                [],
                dedent("""\
                --
                -- depends: 20210101_01_rando-commit

                -- migrate: apply

                -- migrate: rollback

                """),
            ),
            (
                ["--py"],
                dedent('''\
                """

                """

                __depends__ = ["20210101_01_rando-commit"]


                async def apply(db):
                    ...


                async def rollback(db):
                    ...
                '''),
            ),
        ],
    )
    def test_dependency_detected_file_written(
        self,
        monkeypatch,
        cli_runner,
        cwd,
        migration_file_factory,
        extra_args,
        expected,
    ):
        migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        monkeypatch.setattr(cli, "make_file", mock.Mock(return_value=cwd / "new_file.sql"))
        monkeypatch.setattr(cli.subprocess, "call", mock.Mock())
        monkeypatch.setattr(cli.Path, "lstat", mock.Mock(side_effect=[mock.Mock(), mock.Mock()]))

        result = cli_runner.invoke(["new", *extra_args])

        assert result.exit_code == 0, result.output
        with (cwd / "new_file.sql").open() as f:
            assert f.read() == expected

    @pytest.mark.usefixtures("pyproject")
    @pytest.mark.parametrize(
        ("extra_args", "expected"),
        [
            (
                [],
                dedent("""\
                --
                -- depends: 20210101_01_rando-commit 20210101_02_rando-commit

                -- migrate: apply

                -- migrate: rollback

                """),
            ),
            (
                ["--py"],
                dedent('''\
                """

                """

                __depends__ = ["20210101_01_rando-commit", "20210101_02_rando-commit"]


                async def apply(db):
                    ...


                async def rollback(db):
                    ...
                '''),
            ),
        ],
    )
    def test_multiple_head_dependencies_detected_file_written(
        self,
        monkeypatch,
        cli_runner,
        cwd,
        migration_file_factory,
        extra_args,
        expected,
    ):
        migration_file_factory(
            "20200101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20200101_01_rando-commit

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        migration_file_factory(
            "20210101_02_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20200101_01_rando-commit

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        monkeypatch.setattr(cli, "make_file", mock.Mock(return_value=cwd / "new_file.sql"))
        monkeypatch.setattr(cli.subprocess, "call", mock.Mock())
        monkeypatch.setattr(cli.Path, "lstat", mock.Mock(side_effect=[mock.Mock(), mock.Mock()]))

        result = cli_runner.invoke(["new", *extra_args])

        assert result.exit_code == 0, result.output
        with (cwd / "new_file.sql").open() as f:
            assert f.read() == expected

    @pytest.mark.usefixtures("pyproject")
    def test_load_failed_quit(self, monkeypatch, cli_runner):
        monkeypatch.setattr(cli.subprocess, "call", mock.Mock())
        monkeypatch.setattr(cli.Path, "lstat", mock.Mock(side_effect=[mock.Mock(), mock.Mock()]))
        monkeypatch.setattr(cli.Migration, "load", mock.Mock(side_effect=Exception))

        result = cli_runner.invoke(["new"], input="q\n")

        assert result.exit_code == 0, result.output
        cli_runner.assert_output(
            dedent("""\
            Error loading migration.
            Retry editing? [Ynqh]: q
            """),
        )

    @pytest.mark.usefixtures("pyproject")
    def test_retry_ignores_invalid_input(self, monkeypatch, cli_runner):
        monkeypatch.setattr(cli.subprocess, "call", mock.Mock())
        monkeypatch.setattr(cli.Path, "lstat", mock.Mock(side_effect=[mock.Mock(), mock.Mock()]))
        monkeypatch.setattr(cli.Migration, "load", mock.Mock(side_effect=Exception))

        result = cli_runner.invoke(["new"], input="a\nb\nq\n")

        assert result.exit_code == 0, result.output
        cli_runner.assert_output(
            dedent("""\
            Error loading migration.
            Retry editing? [Ynqh]: a
            Retry editing? [Ynqh]: b
            Retry editing? [Ynqh]: q
            """),
        )

    @pytest.mark.usefixtures("pyproject")
    def test_retry_help(self, monkeypatch, cli_runner):
        monkeypatch.setattr(cli.subprocess, "call", mock.Mock())
        monkeypatch.setattr(cli.Path, "lstat", mock.Mock(side_effect=[mock.Mock(), mock.Mock()]))
        monkeypatch.setattr(cli.Migration, "load", mock.Mock(side_effect=Exception))

        result = cli_runner.invoke(["new"], input="h\nq\n")

        assert result.exit_code == 0, result.output
        cli_runner.assert_output(
            dedent("""\
            Error loading migration.
            Retry editing? [Ynqh]: h
            y: reopen the migration file in your editor
            n: save the migration as-is, without re-editing
            q: quit without saving the migration
            h: show this help

            Retry editing? [Ynqh]: q
            """),
        )

    @pytest.mark.usefixtures("pyproject")
    def test_load_failed_retry_and_exit(self, monkeypatch, cli_runner, cwd):
        monkeypatch.setattr(cli, "make_file", mock.Mock(return_value=cwd / "new_file.py"))
        monkeypatch.setattr(cli.subprocess, "call", mock.Mock())
        monkeypatch.setattr(cli.Path, "lstat", mock.Mock(side_effect=[mock.Mock(), mock.Mock(), mock.Mock()]))
        monkeypatch.setattr(cli.Migration, "load", mock.Mock(side_effect=Exception))

        result = cli_runner.invoke(["new"], input="y\nn\n")

        assert result.exit_code == 0, result.output
        cli_runner.assert_output(
            dedent("""\
            Error loading migration.
            Retry editing? [Ynqh]: y
            Error loading migration.
            Retry editing? [Ynqh]: n
            Created file: new_file.py
            """),
        )

    @pytest.mark.usefixtures("pyproject")
    def test_load_failed_default_and_exit(self, monkeypatch, cli_runner, cwd):
        monkeypatch.setattr(cli, "make_file", mock.Mock(return_value=cwd / "new_file.py"))
        monkeypatch.setattr(cli.subprocess, "call", mock.Mock())
        monkeypatch.setattr(cli.Path, "lstat", mock.Mock(side_effect=[mock.Mock(), mock.Mock(), mock.Mock()]))
        monkeypatch.setattr(cli.Migration, "load", mock.Mock(side_effect=Exception))

        result = cli_runner.invoke(["new"], input="\nn\n")

        assert result.exit_code == 0, result.output
        cli_runner.assert_output(
            dedent("""\
            Error loading migration.
            Retry editing? [Ynqh]:
            Error loading migration.
            Retry editing? [Ynqh]: n
            Created file: new_file.py
            """),
        )


class TestHistory:
    @pytest.mark.usefixtures("migrations", "pyproject")
    def test_no_migrations(self, cli_runner):
        result = cli_runner.invoke(["history"])
        assert result.exit_code == 0, result.output
        cli_runner.assert_output(
            dedent("""\
            STATUS    ID    FORMAT
            --------  ----  --------
            """),
        )

    @pytest.mark.usefixtures("migrations", "pyproject")
    def test_migrations_not_applied(self, migration_file_factory, cli_runner):
        migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_02_rando-commit

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        migration_file_factory(
            "20210101_02_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        result = cli_runner.invoke(["history"])
        assert result.exit_code == 0, result.output
        cli_runner.assert_output(
            dedent("""\
            STATUS    ID                        FORMAT
            --------  ------------------------  --------
            U         20210101_02_rando-commit  sql
            U         20210101_01_rando-commit  sql
            """),
        )

    @pytest.mark.usefixtures("migrations", "pyproject_no_database")
    def test_migrations_missing_config(self, migration_file_factory, cli_runner):
        migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_02_rando-commit

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        migration_file_factory(
            "20210101_02_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        result = cli_runner.invoke(["history", "-v"])
        assert result.exit_code == 0, result.output
        cli_runner.assert_output(
            dedent("""\
            Database connection can not be established, migration status can not be determined.
            STATUS    ID                        FORMAT
            --------  ------------------------  --------
            U         20210101_02_rando-commit  sql
            U         20210101_01_rando-commit  sql
            """),
        )

    @pytest.mark.usefixtures("migrations", "pyproject_no_database")
    def test_migrations_missing_config_manual_override(
        self,
        migration_file_factory,
        cli_runner,
        postgres_dsn,
    ):
        migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_02_rando-commit

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        migration_file_factory(
            "20210101_02_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        result = cli_runner.invoke(["history", "-v", "--database", postgres_dsn])
        assert result.exit_code == 0, result.output
        cli_runner.assert_output(
            dedent("""\
            STATUS    ID                        FORMAT
            --------  ------------------------  --------
            U         20210101_02_rando-commit  sql
            U         20210101_01_rando-commit  sql
            """),
        )

    @pytest.mark.usefixtures("migrations", "pyproject")
    async def test_migrations_partial_applied(self, cli_runner, migration_file_factory, db_session):
        await sql.migration_applied(db_session, "20210101_01_rando-commit", "hash", schema_name="public")
        migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        migration_file_factory(
            "20210101_02_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_01_rando-commit

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        result = cli_runner.invoke(["history"])
        assert result.exit_code == 0, result.output
        cli_runner.assert_output(
            dedent("""\
            STATUS    ID                        FORMAT
            --------  ------------------------  --------
            A         20210101_01_rando-commit  sql
            U         20210101_02_rando-commit  sql
            """),
        )

    @pytest.mark.usefixtures("migrations", "pyproject")
    async def test_migrations_partial_applied_only_unapplied(self, cli_runner, migration_file_factory, db_session):
        await sql.migration_applied(db_session, "20210101_01_rando-commit", "hash", schema_name="public")
        migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        migration_file_factory(
            "20210101_02_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_01_rando-commit

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        result = cli_runner.invoke(["history", "--unapplied"])
        assert result.exit_code == 0, result.output
        cli_runner.assert_output(
            dedent("""\
            STATUS    ID                        FORMAT
            --------  ------------------------  --------
            U         20210101_02_rando-commit  sql
            """),
        )

    @pytest.mark.usefixtures("migrations", "pyproject")
    async def test_migrations_partial_applied_unapplied_simple(self, cli_runner, migration_file_factory, db_session):
        await sql.migration_applied(db_session, "20210101_01_rando-commit", "hash", schema_name="public")
        migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        migration_file_factory(
            "20210101_02_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_01_rando-commit

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        result = cli_runner.invoke(["history", "--unapplied", "--simple"])
        assert result.exit_code == 0, result.output
        cli_runner.assert_output(
            dedent("""\
            U 20210101_02_rando-commit sql
            """),
        )


class Base:
    async def assert_tables(self, db_session, tables):
        stmt = """
        SELECT schemaname, tablename
        FROM pg_tables
        WHERE  schemaname in ('public', 'pogo')
        ORDER BY schemaname, tablename
        """
        results = await db_session.fetch(stmt)

        assert [
            f"{r['schemaname']}.{r['tablename']}" for r in results if not r["tablename"].startswith("_pogo")
        ] == tables

    async def assert_schemas(self, db_session, schemas):
        stmt = """
        SELECT schema_name
        FROM information_schema.schemata
        WHERE  schema_name NOT IN ('pg_toast', 'pg_catalog', 'information_schema')
        ORDER BY schema_name
        """
        results = await db_session.fetch(stmt)

        assert [r["schema_name"] for r in results] == schemas


class TestApply(Base):
    @pytest.mark.usefixtures("migrations", "pyproject_no_database")
    async def test_apply_success_missing_config(self, cli_runner, migration_file_factory):
        migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            CREATE TABLE table_one()
            -- migrate: rollback
            """),
        )
        migration_file_factory(
            "20210101_02_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_01_rando-commit

            -- migrate: apply
            CREATE TABLE table_two()
            -- migrate: rollback
            """),
        )
        result = cli_runner.invoke(["apply", "-v"])
        assert result.exit_code == 1, result.output
        cli_runner.assert_output(
            dedent("""\
            Required config `database_config` is not set.
            """),
        )

    @pytest.mark.usefixtures("migrations", "pyproject_no_database")
    async def test_apply_success_missing_config_manual_override(
        self,
        cli_runner,
        migration_file_factory,
        db_session,
        postgres_dsn,
    ):
        migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            CREATE TABLE table_one()
            -- migrate: rollback
            """),
        )
        migration_file_factory(
            "20210101_02_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_01_rando-commit

            -- migrate: apply
            CREATE TABLE table_two()
            -- migrate: rollback
            """),
        )
        result = cli_runner.invoke(["apply", "-v", "--database", postgres_dsn])
        assert result.exit_code == 0, result.output
        cli_runner.assert_output(
            dedent("""\
            Applying 20210101_01_rando-commit
            Applying 20210101_02_rando-commit
            """),
        )
        await self.assert_tables(db_session, ["public.table_one", "public.table_two"])

    @pytest.mark.usefixtures("migrations", "pyproject")
    async def test_apply_success(self, cli_runner, migration_file_factory, db_session):
        migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            CREATE TABLE table_one()
            -- migrate: rollback
            """),
        )
        migration_file_factory(
            "20210101_02_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_01_rando-commit

            -- migrate: apply
            CREATE TABLE table_two()
            -- migrate: rollback
            """),
        )
        result = cli_runner.invoke(["apply", "-v"])
        assert result.exit_code == 0, result.output
        cli_runner.assert_output(
            dedent("""\
            Applying 20210101_01_rando-commit
            Applying 20210101_02_rando-commit
            """),
        )
        await self.assert_tables(db_session, ["public.table_one", "public.table_two"])

    @pytest.mark.usefixtures("migrations", "pyproject")
    async def test_apply_success_multiple_schemas(self, cli_runner, migration_file_factory, db_session):
        migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            CREATE TABLE table_one()
            -- migrate: rollback
            """),
        )
        migration_file_factory(
            "20210101_02_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_01_rando-commit

            -- migrate: apply
            CREATE TABLE table_two()
            -- migrate: rollback
            """),
        )
        result = cli_runner.invoke(["apply", "-v"])
        assert result.exit_code == 0, result.output
        cli_runner.assert_output(
            dedent("""\
            Applying 20210101_01_rando-commit
            Applying 20210101_02_rando-commit
            """),
        )
        result = cli_runner.invoke(["apply", "-v", "--schema", "pogo", "--create-schema"])
        assert result.exit_code == 0, result.output
        cli_runner.assert_output(
            dedent("""\
            Applying 20210101_01_rando-commit
            Applying 20210101_02_rando-commit
            """),
        )
        await self.assert_tables(
            db_session,
            ["pogo.table_one", "pogo.table_two", "public.table_one", "public.table_two"],
        )

    @pytest.mark.usefixtures("migrations", "pyproject")
    async def test_apply_failure(self, cli_runner, migration_file_factory, db_session):
        migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            CREATE TABLE table_one()
            -- migrate: rollback
            """),
        )
        migration_file_factory(
            "20210101_02_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_01_rando-commit

            -- migrate: apply
            CREATE TABLE table_two
            -- migrate: rollback
            """),
        )
        result = cli_runner.invoke(["apply", "-v"])
        assert result.exit_code == 1, result.output
        cli_runner.assert_output(
            dedent("""\
            Applying 20210101_01_rando-commit
            Applying 20210101_02_rando-commit
            Failed to apply 20210101_02_rando-commit
            """),
        )

        await self.assert_tables(db_session, ["public.table_one"])

    @pytest.mark.usefixtures("migrations", "pyproject")
    def test_apply_failure_verbose(self, cli_runner, migration_file_factory):
        migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            CREATE TABLE table_one()
            -- migrate: rollback
            """),
        )
        migration_file_factory(
            "20210101_02_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_01_rando-commit

            -- migrate: apply
            CREATE TABLE table_one()
            -- migrate: rollback
            """),
        )
        result = cli_runner.invoke(["apply", "-vvv"])
        assert result.exit_code == 1, result.output
        assert 'DuplicateTableError: relation "table_one" already exists' in result.output


class TestRollback(Base):
    @pytest.mark.usefixtures("migrations", "pyproject_no_database")
    async def test_rollback_missing_configuration(self, cli_runner, migration_file_factory, db_session):
        await sql.migration_applied(db_session, "20210101_01_rando-commit", "hash", schema_name="public")
        await sql.migration_applied(db_session, "20210101_02_rando-commit", "hash2", schema_name="public")
        await db_session.execute("create table table_one();create table table_two()")
        migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            -- migrate: rollback
            DROP TABLE table_one;
            """),
        )
        migration_file_factory(
            "20210101_02_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_01_rando-commit

            -- migrate: apply
            -- migrate: rollback
            DROP TABLE table_two;
            """),
        )
        result = cli_runner.invoke(["rollback", "--count", "-1", "-v"])
        assert result.exit_code == 1, result.output
        cli_runner.assert_output(
            dedent("""\
            Required config `database_config` is not set.
            """),
        )

    @pytest.mark.usefixtures("migrations", "pyproject_no_database")
    async def test_rollback_missing_configuration_manual_override(
        self,
        cli_runner,
        migration_file_factory,
        db_session,
        postgres_dsn,
    ):
        await sql.migration_applied(db_session, "20210101_01_rando-commit", "hash", schema_name="public")
        await sql.migration_applied(db_session, "20210101_02_rando-commit", "hash2", schema_name="public")
        await db_session.execute("create table table_one();create table table_two()")
        migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            -- migrate: rollback
            DROP TABLE table_one;
            """),
        )
        migration_file_factory(
            "20210101_02_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_01_rando-commit

            -- migrate: apply
            -- migrate: rollback
            DROP TABLE table_two;
            """),
        )
        result = cli_runner.invoke(["rollback", "--count", "-1", "-v", "--database", postgres_dsn])
        assert result.exit_code == 0, result.output
        cli_runner.assert_output(
            dedent("""\
            Rolling back 20210101_02_rando-commit
            Rolling back 20210101_01_rando-commit
            """),
        )
        await self.assert_tables(db_session, [])

    @pytest.mark.usefixtures("migrations", "pyproject")
    async def test_rollback_success(self, cli_runner, migration_file_factory, db_session):
        await sql.migration_applied(db_session, "20210101_01_rando-commit", "hash", schema_name="public")
        await sql.migration_applied(db_session, "20210101_02_rando-commit", "hash2", schema_name="public")
        await db_session.execute("create table table_one();create table table_two()")
        migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            -- migrate: rollback
            DROP TABLE table_one;
            """),
        )
        migration_file_factory(
            "20210101_02_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_01_rando-commit

            -- migrate: apply
            -- migrate: rollback
            DROP TABLE table_two;
            """),
        )
        result = cli_runner.invoke(["rollback", "--count", "-1", "-v"])
        assert result.exit_code == 0, result.output
        cli_runner.assert_output(
            dedent("""\
            Rolling back 20210101_02_rando-commit
            Rolling back 20210101_01_rando-commit
            """),
        )
        await self.assert_tables(db_session, [])

    @pytest.mark.usefixtures("migrations", "pyproject")
    async def test_rollback_partial_success(self, cli_runner, migration_file_factory, db_session):
        await sql.migration_applied(db_session, "20210101_01_rando-commit", "hash", schema_name="public")
        await sql.migration_applied(db_session, "20210101_02_rando-commit", "hash2", schema_name="public")
        await db_session.execute("create table table_one();create table table_two()")
        migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            -- migrate: rollback
            DROP TABLE table_one;
            """),
        )
        migration_file_factory(
            "20210101_02_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_01_rando-commit

            -- migrate: apply
            -- migrate: rollback
            DROP TABLE table_two;
            """),
        )
        result = cli_runner.invoke(["rollback", "--count", "1", "-v"])
        assert result.exit_code == 0, result.output
        cli_runner.assert_output(
            dedent("""\
            Rolling back 20210101_02_rando-commit
            """),
        )
        await self.assert_tables(db_session, ["public.table_one"])

    @pytest.mark.usefixtures("migrations", "pyproject")
    async def test_rollback_drop_schema(self, cli_runner, migration_file_factory, db_session):
        migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            CREATE TABLE table_one()
            -- migrate: rollback
            DROP TABLE table_one;
            """),
        )
        migration_file_factory(
            "20210101_02_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_01_rando-commit

            -- migrate: apply
            CREATE TABLE table_two()
            -- migrate: rollback
            DROP TABLE table_two;
            """),
        )
        cli_runner.invoke(["apply", "--schema", "pogo", "--create-schema"])
        result = cli_runner.invoke(["rollback", "--count", "-1", "-v", "--schema", "pogo", "--drop-schema"])
        assert result.exit_code == 0, result.output
        cli_runner.assert_output(
            dedent("""\
            Rolling back 20210101_02_rando-commit
            Rolling back 20210101_01_rando-commit
            """),
        )
        await self.assert_tables(db_session, [])
        await self.assert_schemas(db_session, ["public"])

    @pytest.mark.usefixtures("migrations", "pyproject")
    async def test_rollback_partial_wont_drop_schema(self, cli_runner, migration_file_factory, db_session):
        migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            CREATE TABLE table_one()
            -- migrate: rollback
            DROP TABLE table_one;
            """),
        )
        migration_file_factory(
            "20210101_02_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_01_rando-commit

            -- migrate: apply
            CREATE TABLE table_two()
            -- migrate: rollback
            DROP TABLE table_two;
            """),
        )
        cli_runner.invoke(["apply", "--schema", "pogo", "--create-schema"])
        result = cli_runner.invoke(["rollback", "--count", "1", "-v", "--schema", "pogo", "--drop-schema"])
        assert result.exit_code == 1, result.output
        cli_runner.assert_output(
            dedent("""\
            Rolling back 20210101_02_rando-commit
            migrations still exist, can't drop schema.
            """),
        )
        await self.assert_tables(db_session, ["pogo.table_one"])
        await self.assert_schemas(db_session, ["pogo", "public"])

    @pytest.mark.usefixtures("migrations", "pyproject")
    async def test_rollback_partial_success_multi_schema(self, cli_runner, migration_file_factory, db_session):
        await sql.migration_applied(db_session, "20210101_01_rando-commit", "hash", schema_name="public")
        await sql.migration_applied(db_session, "20210101_02_rando-commit", "hash2", schema_name="public")
        await db_session.execute("create table table_one();create table table_two()")
        migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            CREATE TABLE table_one()
            -- migrate: rollback
            DROP TABLE table_one;
            """),
        )
        migration_file_factory(
            "20210101_02_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_01_rando-commit

            -- migrate: apply
            CREATE TABLE table_two()
            -- migrate: rollback
            DROP TABLE table_two;
            """),
        )
        cli_runner.invoke(["apply"])
        cli_runner.invoke(["apply", "--schema", "pogo", "--create-schema"])
        result = cli_runner.invoke(["rollback", "--count", "1", "-v", "--schema", "pogo"])
        assert result.exit_code == 0, result.output
        cli_runner.assert_output(
            dedent("""\
            Rolling back 20210101_02_rando-commit
            """),
        )
        await self.assert_tables(db_session, ["pogo.table_one", "public.table_one", "public.table_two"])

    @pytest.mark.usefixtures("migrations", "pyproject")
    async def test_rollback_failure(self, cli_runner, migration_file_factory, db_session):
        await sql.migration_applied(db_session, "20210101_01_rando-commit", "hash", schema_name="public")
        await db_session.execute("create table table_one();create table table_two()")
        migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            -- migrate: rollback
            DROP TABLE table_one;
            DROP TABLE table_two;
            DROP TABLE table_three;
            """),
        )
        result = cli_runner.invoke(["rollback", "-v"])
        assert result.exit_code == 1, result.output
        cli_runner.assert_output(
            dedent("""\
            Rolling back 20210101_01_rando-commit
            Failed to rollback 20210101_01_rando-commit
            """),
        )

        await self.assert_tables(db_session, ["public.table_one", "public.table_two"])

    @pytest.mark.usefixtures("migrations", "pyproject")
    async def test_rollback_failure_verbose(self, cli_runner, migration_file_factory, db_session):
        await sql.migration_applied(db_session, "20210101_01_rando-commit", "hash", schema_name="public")
        migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            -- migrate: rollback
            DROP TABLE table_one;
            """),
        )
        result = cli_runner.invoke(["rollback", "-vvv"])
        assert result.exit_code == 1, result.output
        assert 'UndefinedTableError: table "table_one" does not exist' in result.output


class TestValidate:
    @pytest.mark.usefixtures("migrations", "pyproject")
    async def test_validate_clean(self, cli_runner, migration_file_factory):
        migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            CREATE TABLE table_one()
            -- migrate: rollback
            """),
        )
        migration_file_factory(
            "20210101_02_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_01_rando-commit

            -- migrate: apply
            CREATE TABLE table_two()
            -- migrate: rollback
            """),
        )
        result = cli_runner.invoke(["validate", "-v"])
        assert result.exit_code == 0, result.output
        cli_runner.assert_output("")

    @pytest.mark.usefixtures("migrations", "pyproject")
    async def test_validate_invalid_sql(self, cli_runner, migration_file_factory):
        migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            CREATE TABLE table_one();
            -- migrate: rollback
            DROP TABLE;
            """),
        )
        migration_file_factory(
            "20210101_02_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_01_rando-commit

            -- migrate: apply
            CREATE INDEX foo;
            -- migrate: rollback
            DROP INDEX;
            DROP AGGREGATE lock;
            """),
        )
        result = cli_runner.invoke(["validate", "-v"])
        assert result.exit_code == 0, result.output
        cli_runner.assert_output(
            dedent("""\
            20210101_01_rando-commit: Expected table name but got None. Line: 1, Column: 10
            DROP TABLE;
            20210101_02_rando-commit: Expected table name but got None. Line: 1, Column: 16
            CREATE INDEX foo;
            20210101_02_rando-commit: Expected table name but got None. Line: 1, Column: 10
            DROP INDEX;
            sqlglot failed to parse, falling back to sqlparse.
            Can not extract table from DDL statement in migration 20210101_02_rando-commit, check that table name is not a reserved word.
            DROP AGGREGATE lock;
            """),
        )

    @pytest.mark.usefixtures("migrations", "pyproject")
    def test_validate_checks_py(self, cli_runner, migration_file_factory):
        migration_file_factory(
            "20210101_01_rando-commit",
            "py",
            dedent('''
            """
            second migration
            """
            __depends__ = []
            __transaction__ = False

            async def apply(db):
                await db.execute("CREATE TABLE one();")
                await db.execute("CREATE INDEX foo;")

            async def rollback(db):
                await db.execute("DROP INDEX;")
                await db.execute("DROP TABLE;")
                await db.execute("DROP AGGREGATE lock;")
            '''),
        )
        result = cli_runner.invoke(["validate", "-v"])
        assert result.exit_code == 0, result.output
        cli_runner.assert_output(
            dedent("""\
            20210101_01_rando-commit: Expected table name but got None. Line: 1, Column: 16
            CREATE INDEX foo;
            20210101_01_rando-commit: Expected table name but got None. Line: 1, Column: 10
            DROP INDEX;
            20210101_01_rando-commit: Expected table name but got None. Line: 1, Column: 10
            DROP TABLE;
            sqlglot failed to parse, falling back to sqlparse.
            Can not extract table from DDL statement in migration 20210101_01_rando-commit, check that table name is not a reserved word.
            DROP AGGREGATE lock;
            """),
        )

    @pytest.mark.usefixtures("migrations", "pyproject")
    def test_validate_py_parametrized(self, cli_runner, migration_file_factory):
        migration_file_factory(
            "20210101_01_abcde-first-migration",
            "py",
            dedent('''
            """
            second migration
            """
            __depends__ = []
            __transaction__ = False

            async def apply(db):
                a = "1"
                await db.execute('UPDATE "table" SET col = $1', a)

            async def rollback(db):
                await db.execute(query='DROP TABLE "table";')
            '''),
        )
        result = cli_runner.invoke(["validate", "-v"])
        assert result.exit_code == 0, result.output
        cli_runner.assert_output(
            dedent("""\
            """),
        )

    @pytest.mark.usefixtures("migrations", "pyproject")
    def test_validate_skips_py_error(self, cli_runner, migration_file_factory):
        migration_file_factory(
            "20210101_01_abcde-first-migration",
            "py",
            dedent('''
            """
            second migration
            """
            __depends__ = []
            __transaction__ = False

            async def apply(db):
                raise Exception

            async def rollback(db):
                raise Exception
            '''),
        )
        result = cli_runner.invoke(["validate", "-v"])
        assert result.exit_code == 0, result.output
        cli_runner.assert_output(
            dedent("""\
            Can't validate python migration 20210101_01_abcde-first-migration (apply), skipping...
            Can't validate python migration 20210101_01_abcde-first-migration (rollback), skipping...
            """),
        )


class TestMark:
    @pytest.mark.usefixtures("migrations", "pyproject")
    def test_mark_no_migrations(self, cli_runner):
        result = cli_runner.invoke(["mark"])
        assert result.exit_code == 0, result.output
        cli_runner.assert_output(
            "",
        )

    @pytest.mark.usefixtures("migrations", "pyproject_no_database")
    async def test_mark_missing_config(self, cli_runner, migration_file_factory, db_session):
        await sql.migration_applied(db_session, "20210101_01_rando-commit", "hash", schema_name="public")
        migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        migration_file_factory(
            "20210101_02_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_01_rando-commit

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        migration_file_factory(
            "20210101_03_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_02_rando-commit

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        result = cli_runner.invoke(["mark"], input="y\nn\n")
        assert result.exit_code == 1, result.output
        cli_runner.assert_output(
            dedent("""\
            Required config `database_config` is not set.
            """),
        )

    @pytest.mark.usefixtures("migrations", "pyproject_no_database")
    async def test_mark_missing_config_manual_override(
        self,
        cli_runner,
        migration_file_factory,
        db_session,
        postgres_dsn,
    ):
        await sql.migration_applied(db_session, "20210101_01_rando-commit", "hash", schema_name="public")
        migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        migration_file_factory(
            "20210101_02_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_01_rando-commit

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        migration_file_factory(
            "20210101_03_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_02_rando-commit

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        result = cli_runner.invoke(["mark", "--database", postgres_dsn], input="y\nn\n")
        assert result.exit_code == 0, result.output
        cli_runner.assert_output(
            dedent("""\
            Mark 20210101_02_rando-commit as applied? [y/N]: y
            Mark 20210101_03_rando-commit as applied? [y/N]: n
            """),
        )
        applied_migrations = await sql.get_applied_migrations(db_session, schema_name="public")
        assert applied_migrations == {"20210101_01_rando-commit", "20210101_02_rando-commit"}

    @pytest.mark.usefixtures("migrations", "pyproject")
    async def test_mark_migrations_applied(self, cli_runner, migration_file_factory, db_session):
        await sql.migration_applied(db_session, "20210101_01_rando-commit", "hash", schema_name="public")
        migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        migration_file_factory(
            "20210101_02_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_01_rando-commit

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        migration_file_factory(
            "20210101_03_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_02_rando-commit

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        result = cli_runner.invoke(["mark"], input="y\nn\n")
        assert result.exit_code == 0, result.output
        cli_runner.assert_output(
            dedent("""\
            Mark 20210101_02_rando-commit as applied? [y/N]: y
            Mark 20210101_03_rando-commit as applied? [y/N]: n
            """),
        )
        applied_migrations = await sql.get_applied_migrations(db_session, schema_name="public")
        assert applied_migrations == {"20210101_01_rando-commit", "20210101_02_rando-commit"}

    @pytest.mark.usefixtures("migrations", "pyproject")
    async def test_mark_migration_applied(self, cli_runner, migration_file_factory, db_session):
        await sql.migration_applied(db_session, "20210101_01_rando-commit", "hash", schema_name="public")
        migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        migration_file_factory(
            "20210101_02_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_01_rando-commit

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        migration_file_factory(
            "20210101_03_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_02_rando-commit

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        result = cli_runner.invoke(["mark", "-m", "20210101_02_rando-commit"], input="y\n")
        assert result.exit_code == 0, result.output
        cli_runner.assert_output(
            dedent("""\
            Mark 20210101_02_rando-commit as applied? [y/N]: y
            """),
        )
        applied_migrations = await sql.get_applied_migrations(db_session, schema_name="public")
        assert applied_migrations == {"20210101_01_rando-commit", "20210101_02_rando-commit"}

    @pytest.mark.usefixtures("migrations", "pyproject")
    async def test_mark_migrations_non_interactive(self, cli_runner, migration_file_factory, db_session):
        await sql.migration_applied(db_session, "20210101_01_rando-commit", "hash", schema_name="public")
        migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        migration_file_factory(
            "20210101_02_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_01_rando-commit

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        migration_file_factory(
            "20210101_03_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_02_rando-commit

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        result = cli_runner.invoke(["mark", "--no-interactive"])
        assert result.exit_code == 0, result.output
        cli_runner.assert_output(
            dedent(""),
        )
        applied_migrations = await sql.get_applied_migrations(db_session, schema_name="public")
        assert applied_migrations == {
            "20210101_01_rando-commit",
            "20210101_02_rando-commit",
            "20210101_03_rando-commit",
        }


class TestUnMark:
    @pytest.mark.usefixtures("migrations", "pyproject")
    def test_unmark_no_migrations(self, cli_runner):
        result = cli_runner.invoke(["unmark"])
        assert result.exit_code == 0, result.output
        cli_runner.assert_output(
            "",
        )

    @pytest.mark.usefixtures("migrations", "pyproject_no_database")
    async def test_unmark_missing_config(self, cli_runner, migration_file_factory, db_session):
        await sql.migration_applied(db_session, "20210101_01_rando-commit", "hash", schema_name="public")
        await sql.migration_applied(db_session, "20210101_02_rando-commit", "hash2", schema_name="public")
        migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        migration_file_factory(
            "20210101_02_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_01_rando-commit

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        migration_file_factory(
            "20210101_03_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_02_rando-commit

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        result = cli_runner.invoke(["unmark"], input="y\nn\n")
        assert result.exit_code == 1, result.output
        cli_runner.assert_output(
            dedent("""\
            Required config `database_config` is not set.
            """),
        )

    @pytest.mark.usefixtures("migrations", "pyproject_no_database")
    async def test_unmark_missing_config_manual_override(
        self,
        cli_runner,
        migration_file_factory,
        db_session,
        postgres_dsn,
    ):
        await sql.migration_applied(db_session, "20210101_01_rando-commit", "hash", schema_name="public")
        await sql.migration_applied(db_session, "20210101_02_rando-commit", "hash2", schema_name="public")
        migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        migration_file_factory(
            "20210101_02_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_01_rando-commit

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        migration_file_factory(
            "20210101_03_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_02_rando-commit

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        result = cli_runner.invoke(["unmark", "--database", postgres_dsn], input="y\nn\n")
        assert result.exit_code == 0, result.output
        cli_runner.assert_output(
            dedent("""\
            Unmark 20210101_02_rando-commit as applied? [y/N]: y
            Unmark 20210101_01_rando-commit as applied? [y/N]: n
            """),
        )
        applied_migrations = await sql.get_applied_migrations(db_session, schema_name="public")
        assert applied_migrations == {"20210101_01_rando-commit"}

    @pytest.mark.usefixtures("migrations", "pyproject")
    async def test_unmark_migrations(self, cli_runner, migration_file_factory, db_session):
        await sql.migration_applied(db_session, "20210101_01_rando-commit", "hash", schema_name="public")
        await sql.migration_applied(db_session, "20210101_02_rando-commit", "hash2", schema_name="public")
        migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        migration_file_factory(
            "20210101_02_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_01_rando-commit

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        migration_file_factory(
            "20210101_03_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_02_rando-commit

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        result = cli_runner.invoke(["unmark"], input="y\nn\n")
        assert result.exit_code == 0, result.output
        cli_runner.assert_output(
            dedent("""\
            Unmark 20210101_02_rando-commit as applied? [y/N]: y
            Unmark 20210101_01_rando-commit as applied? [y/N]: n
            """),
        )
        applied_migrations = await sql.get_applied_migrations(db_session, schema_name="public")
        assert applied_migrations == {"20210101_01_rando-commit"}

    @pytest.mark.usefixtures("migrations", "pyproject")
    async def test_unmark_migration(self, cli_runner, migration_file_factory, db_session):
        await sql.migration_applied(db_session, "20210101_01_rando-commit", "hash", schema_name="public")
        await sql.migration_applied(db_session, "20210101_02_rando-commit", "hash2", schema_name="public")
        migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        migration_file_factory(
            "20210101_02_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_01_rando-commit

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        migration_file_factory(
            "20210101_03_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_02_rando-commit

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        result = cli_runner.invoke(["unmark", "-m", "20210101_02_rando-commit"], input="y\n")
        assert result.exit_code == 0, result.output
        cli_runner.assert_output(
            dedent("""\
            Unmark 20210101_02_rando-commit as applied? [y/N]: y
            """),
        )
        applied_migrations = await sql.get_applied_migrations(db_session, schema_name="public")
        assert applied_migrations == {"20210101_01_rando-commit"}


class TestMigrateYoyo:
    @pytest.mark.usefixtures("migrations", "pyproject")
    async def test_skip_files(self, migration_file_factory, cli_runner, db_session):
        await db_session.execute("""
        create table _yoyo_migration (
            migration_hash varchar(64),
            migration_id varchar(255),
            applied_at_utc timestamp
        );""")
        migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            CREATE TABLE table_one();
            """),
        )
        migration_file_factory(
            "20210101_01_rando-commit",
            "rollback.sql",
            dedent("""
            -- commit
            -- depends:
            DROP TABLE table_one;
            """),
        )
        migration_file_factory(
            "20210101_02_rando-commit2",
            "sql",
            dedent("""
            -- commit2
            -- depends: 20210101_01_rando-commit
            CREATE TABLE table_two()
            """),
        )

        result = cli_runner.invoke(["migrate-yoyo", "--skip-files", "-vvv"])
        assert result.exit_code == 0
        cli_runner.assert_output(
            dedent("""\
            skip-files set, ignoring existing migration files.
            """),
        )

    @pytest.mark.usefixtures("migrations", "pyproject_no_database")
    async def test_missing_config(self, cli_runner, db_session):
        await db_session.execute("""
        create table _yoyo_migration (
            migration_hash varchar(64),
            migration_id varchar(255),
            applied_at_utc timestamp
        );""")
        await sql.migration_applied(db_session, "20210101_01_rando-commit", "hash", schema_name="public")

        result = cli_runner.invoke(["migrate-yoyo", "-v"])
        assert result.exit_code == 1
        cli_runner.assert_output(
            dedent("""\
            Required config `database_config` is not set.
            """),
        )

    @pytest.mark.usefixtures("migrations", "pyproject_no_database")
    async def test_missing_config_manual_override(self, cli_runner, db_session, postgres_dsn):
        await db_session.execute("""
        create table _yoyo_migration (
            migration_hash varchar(64),
            migration_id varchar(255),
            applied_at_utc timestamp
        );""")
        await sql.migration_applied(db_session, "20210101_01_rando-commit", "hash", schema_name="public")

        result = cli_runner.invoke(["migrate-yoyo", "-vvv", "--database", postgres_dsn])
        assert result.exit_code == 0
        cli_runner.assert_output(
            dedent("""\
            migration history exists, skipping yoyo migration.
            """),
        )

    @pytest.mark.usefixtures("migrations", "pyproject")
    async def test_history_already_loaded(self, cli_runner, db_session):
        await db_session.execute("""
        create table _yoyo_migration (
            migration_hash varchar(64),
            migration_id varchar(255),
            applied_at_utc timestamp
        );""")
        await sql.migration_applied(db_session, "20210101_01_rando-commit", "hash", schema_name="public")

        result = cli_runner.invoke(["migrate-yoyo", "-vvv"])
        assert result.exit_code == 0
        cli_runner.assert_output(
            dedent("""\
            migration history exists, skipping yoyo migration.
            """),
        )

    @pytest.mark.usefixtures("migrations", "pyproject")
    async def test_no_yoyo_history(self, cli_runner):
        result = cli_runner.invoke(["migrate-yoyo", "-vvv"])
        assert result.exit_code == 0
        cli_runner.assert_output(
            dedent("""\
            yoyo migration history missing, skipping yoyo migration.
            """),
        )

    @pytest.mark.usefixtures("migrations", "pyproject")
    async def test_sql_files_converted(self, migration_file_factory, cli_runner, db_session):
        await db_session.execute("""
        create table _yoyo_migration (
            migration_hash varchar(64),
            migration_id varchar(255),
            applied_at_utc timestamp
        );""")
        migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            CREATE TABLE table_one();
            """),
        )
        migration_file_factory(
            "20210101_01_rando-commit",
            "rollback.sql",
            dedent("""
            -- commit
            -- depends:
            DROP TABLE table_one;
            """),
        )
        migration_file_factory(
            "20210101_02_rando-commit2",
            "sql",
            dedent("""
            -- commit2
            -- depends: 20210101_01_rando-commit
            CREATE TABLE table_two()
            """),
        )

        result = cli_runner.invoke(["migrate-yoyo"])
        assert result.exit_code == 0
        cli_runner.assert_output(
            dedent("""\
            Converted 'migrations/20210101_01_rando-commit.sql' successfully.
            Converted 'migrations/20210101_02_rando-commit2.sql' successfully.
            """),
        )

    @pytest.mark.usefixtures("migrations", "pyproject")
    async def test_py_files_skipped(self, migration_file_factory, cli_runner, db_session):
        await db_session.execute("""
        create table _yoyo_migration (
            migration_hash varchar(64),
            migration_id varchar(255),
            applied_at_utc timestamp
        );""")
        migration_file_factory(
            "20210101_01_rando-commit",
            "py",
            "ignored",
        )

        result = cli_runner.invoke(["migrate-yoyo"])
        assert result.exit_code == 0, result.output
        cli_runner.assert_output(
            dedent("""\
            Python files can not be migrated reliably, please manually update 'migrations/20210101_01_rando-commit.py'.
            """),
        )


@pytest.mark.usefixtures("pyproject")
class TestRemove:
    def test_migration_removed_from_chain(self, migration_file_factory, cli_runner):
        migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_02_rando-commit

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        mp = migration_file_factory(
            "20210101_02_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            -- migrate: rollback
            """),
        )

        result = cli_runner.invoke(["remove", "20210101_02_rando"])
        assert result.exit_code == 0, result.output

        assert mp.exists() is False

    def test_migration_removed_from_chain_with_backup(self, migration_file_factory, cli_runner):
        migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_02_rando-commit

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        mp = migration_file_factory(
            "20210101_02_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            -- migrate: rollback
            """),
        )

        result = cli_runner.invoke(["remove", "20210101_02_rando", "--backup"])
        assert result.exit_code == 0, result.output

        assert mp.exists() is False
        assert Path(f"{mp}.bak").exists() is True


@pytest.mark.usefixtures("pyproject")
class TestClean:
    def test_bak_files_removed(self, migration_file_factory, cli_runner):
        b1 = migration_file_factory(
            "20210101_01_rando-commit",
            "sql.bak",
            dedent("""
            -- commit
            -- depends: 20210101_02_rando-commit

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        b2 = migration_file_factory(
            "20210101_02_rando-commit",
            "sql.bak",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        mp = migration_file_factory(
            "20210101_02_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            -- migrate: rollback
            """),
        )

        result = cli_runner.invoke(["clean"])
        assert result.exit_code == 0, result.output

        assert b1.exists() is False
        assert b2.exists() is False
        assert mp.exists() is True


@pytest.mark.usefixtures("pyproject")
class TestSquash:
    def test_simple_squash_all_sql(self, migration_file_factory, cli_runner):
        old = migration_file_factory(
            "20210101_01_abcd-first-migration",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            CREATE TABLE one (id INT);
            INSERT INTO one (id) VALUES (1);
            -- migrate: rollback
            DELETE FROM one;
            DROP TABLE one;
            """),
        )
        new = migration_file_factory(
            "20210101_02_efgh-second-migration",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_01_abcd-first-migration

            -- migrate: apply
            CREATE TABLE two (id INT);
            UPDATE one SET id = 2;
            -- migrate: rollback
            UPDATE one SET id = 1;
            DROP TABLE two;
            """),
        )

        result = cli_runner.invoke(["squash", "-vvv"])
        assert result.exit_code == 0, result.output

        assert old.exists() is False
        assert new.read_text() == dedent("""\
        -- commit
        -- depends:

        -- squashed: 20210101_01_abcd-first-migration

        -- migrate: apply

        -- Squash one statements.

        CREATE TABLE one (id INT);

        -- Squash two statements.

        CREATE TABLE two (id INT);

        -- Squash data statements.

        INSERT INTO one (id) VALUES (1);

        UPDATE one SET id = 2;

        -- migrate: rollback

        -- Squash data statements.

        UPDATE one SET id = 1;

        DELETE FROM one;

        -- Squash two statements.

        DROP TABLE two;

        -- Squash one statements.

        DROP TABLE one;
        """)

    def test_apply_sqlglot_reserved_keyword_names_errors_trapped(self, migration_file_factory, cli_runner):
        migration_file_factory(
            "20210101_01_abcd-first-migration",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            CREATE TABLE lock (id INT);
            -- migrate: rollback
            """),
        )

        result = cli_runner.invoke(["squash"])
        assert result.exit_code == 1, result.output

        cli_runner.assert_output(
            "20210101_01_abcd-first-migration: Expected table name but got lock. Line: 1, Column: 17",
        )

    def test_apply_reserved_keyword_names_errors_trapped(self, migration_file_factory, cli_runner):
        migration_file_factory(
            "20210101_01_abcd-first-migration",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            DROP AGGREGATE lock;
            -- migrate: rollback
            """),
        )

        result = cli_runner.invoke(["squash"])
        assert result.exit_code == 1, result.output

        cli_runner.assert_output(
            "Can not extract table from DDL statement in migration 20210101_01_abcd-first-migration",
        )

    def test_rollback_sqlglot_reserved_keyword_names_errors_trapped(self, migration_file_factory, cli_runner):
        migration_file_factory(
            "20210101_01_abcd-first-migration",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            -- migrate: rollback
            CREATE TABLE lock;
            """),
        )

        result = cli_runner.invoke(["squash"])
        assert result.exit_code == 1, result.output

        cli_runner.assert_output(
            "20210101_01_abcd-first-migration: Expected table name but got lock. Line: 1, Column: 17",
        )

    def test_rollback_reserved_keyword_names_errors_trapped(self, migration_file_factory, cli_runner):
        migration_file_factory(
            "20210101_01_abcd-first-migration",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            -- migrate: rollback
            CREATE AGGREGATE lock;
            """),
        )

        result = cli_runner.invoke(["squash"])
        assert result.exit_code == 1, result.output

        cli_runner.assert_output(
            "Can not extract table from DDL statement in migration 20210101_01_abcd-first-migration",
        )

    def test_prompt_update(self, migration_file_factory, cli_runner):
        migration_file_factory(
            "20210101_01_abcd-first-migration",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            CREATE TABLE one (id INT);
            UPDATE one SET id = 1;
            -- migrate: rollback
            DROP TABLE one;
            """),
        )
        new = migration_file_factory(
            "20210101_02_efgh-second-migration",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_01_abcd-first-migration

            -- migrate: apply
            CREATE TABLE two (id INT);
            UPDATE one SET id = 2;
            -- migrate: rollback
            DROP TABLE two;
            """),
        )

        result = cli_runner.invoke(["squash", "--update-prompt"], input="\nn\n")
        assert result.exit_code == 0, result.output

        assert new.read_text() == dedent("""\
        -- commit
        -- depends:

        -- squashed: 20210101_01_abcd-first-migration

        -- migrate: apply

        -- Squash one statements.

        CREATE TABLE one (id INT);

        -- Squash two statements.

        CREATE TABLE two (id INT);

        -- Squash data statements.

        UPDATE one SET id = 1;

        -- migrate: rollback

        -- Squash two statements.

        DROP TABLE two;

        -- Squash one statements.

        DROP TABLE one;
        """)

    def test_python_and_non_transaction_skipped(self, migration_file_factory, cli_runner, migrations):
        migration_file_factory(
            "20210101_01_abcd-first-migration",
            "sql",
            dedent("""
            -- first migration
            -- depends:

            -- migrate: apply
            CREATE TABLE one (id INT);

            -- migrate: rollback
            DROP TABLE one;
            """),
        )
        migration_file_factory(
            "20210101_02_efgh-second-migration",
            "py",
            dedent('''
            """
            second migration
            """
            __depends__ = ["20210101_01_abcd-first-migration"]
            __transaction__ = False

            async def apply(db):
                await db.execute("CREATE TABLE two();")

            async def rollback(db):
                await db.execute("DROP TABLE two;")
            '''),
        )
        migration_file_factory(
            "20210101_03_ijkl-third-migration",
            "sql",
            dedent("""
            -- third migration
            -- depends: 20210101_02_efgh-second-migration

            -- migrate: apply
            CREATE TABLE three (id INT);
            -- migrate: rollback
            DROP TABLE three;
            """),
        )
        new = migration_file_factory(
            "20210101_04_mnop-fourth-migration",
            "sql",
            dedent("""
            -- fourth migration
            -- depends: 20210101_03_ijkl-third-migration

            -- migrate: apply
            CREATE TABLE four (id INT);
            -- migrate: rollback
            DROP TABLE four;
            """),
        )
        migration_file_factory(
            "20210102_01_abcd-fifth-migration",
            "sql",
            dedent("""
            -- fifth migration
            -- depends: 20210101_04_mnop-fourth-migration

            -- transaction: false

            -- migrate: apply
            CREATE TABLE five (id INT);
            -- migrate: rollback
            DROP TABLE five;
            """),
        )

        result = cli_runner.invoke(["squash"])
        assert result.exit_code == 0, result.output

        assert new.read_text() == dedent("""\
        -- fourth migration
        -- depends: 20210101_02_efgh-second-migration

        -- squashed: 20210101_03_ijkl-third-migration

        -- migrate: apply

        -- Squash three statements.

        CREATE TABLE three (id INT);

        -- Squash four statements.

        CREATE TABLE four (id INT);

        -- migrate: rollback

        -- Squash four statements.

        DROP TABLE four;

        -- Squash three statements.

        DROP TABLE three;
        """)

        assert sorted([path.stem for path in migrations.iterdir() if path.suffix in {".py", ".sql"}]) == [
            "20210101_01_abcd-first-migration",
            "20210101_02_efgh-second-migration",
            "20210101_04_mnop-fourth-migration",
            "20210102_01_abcd-fifth-migration",
        ]

    def test_squash_exclusions_honoured(self, migration_file_factory, cli_runner, migrations, pyproject_factory):
        pyproject_factory(
            {
                "migrations": "./migrations",
                "squash": {"exclude": ["20210101_03_ijkl-third-migration"]},
            },
        )
        migration_file_factory(
            "20210101_01_abcd-first-migration",
            "sql",
            dedent("""
            -- first migration
            -- depends:

            -- migrate: apply
            CREATE TABLE one (id INT);

            -- migrate: rollback
            DROP TABLE one;
            """),
        )
        migration_file_factory(
            "20210101_02_efgh-second-migration",
            "sql",
            dedent("""
            -- second migration
            -- depends: 20210101_01_abcd-first-migration

            -- migrate: apply
            CREATE TABLE two();

            -- migrate: rollback
            DROP TABLE two;
            """),
        )
        migration_file_factory(
            "20210101_03_ijkl-third-migration",
            "sql",
            dedent("""
            -- third migration
            -- depends: 20210101_02_efgh-second-migration

            -- migrate: apply
            CREATE TABLE three (id INT);
            -- migrate: rollback
            DROP TABLE three;
            """),
        )
        migration_file_factory(
            "20210101_04_mnop-fourth-migration",
            "sql",
            dedent("""
            -- fourth migration
            -- depends: 20210101_03_ijkl-third-migration

            -- migrate: apply
            CREATE TABLE four (id INT);
            -- migrate: rollback
            DROP TABLE four;
            """),
        )
        migration_file_factory(
            "20210102_01_abcd-fifth-migration",
            "sql",
            dedent("""
            -- fifth migration
            -- depends: 20210101_04_mnop-fourth-migration

            -- migrate: apply
            CREATE TABLE five (id INT);
            -- migrate: rollback
            DROP TABLE five;
            """),
        )

        result = cli_runner.invoke(["squash"])
        assert result.exit_code == 0, result.output

        assert sorted([path.stem for path in migrations.iterdir() if path.suffix in {".py", ".sql"}]) == [
            "20210101_02_efgh-second-migration",
            "20210101_03_ijkl-third-migration",
            "20210102_01_abcd-fifth-migration",
        ]

    def test_skip_initial_file(self, migration_file_factory, cli_runner, migrations):
        migration_file_factory(
            "20210101_01_abcd-first-migration",
            "sql",
            dedent("""
            -- first migration
            -- depends:
            -- transaction: false

            -- migrate: apply
            CREATE TABLE one (id INT);
            -- migrate: rollback
            DROP TABLE one;
            """),
        )

        result = cli_runner.invoke(["squash"])
        assert result.exit_code == 0, result.output

        assert sorted([path.stem for path in migrations.iterdir() if path.suffix in {".py", ".sql"}]) == [
            "20210101_01_abcd-first-migration",
        ]

    def test_view_and_remove_python_file(self, migration_file_factory, cli_runner, migrations):
        migration_file_factory(
            "20210101_01_abcd-first-migration",
            "py",
            dedent('''
            """
            first migration
            """
            __depends__ = []
            __transaction__ = False

            async def apply(db):
                await db.execute("CREATE TABLE two();")

            async def rollback(db):
                await db.execute("DROP TABLE two;")
            '''),
        )

        result = cli_runner.invoke(["squash", "--skip-prompt"], input="y\ny\n")
        assert result.exit_code == 0, result.output
        cli_runner.assert_output(
            dedent('''\
        View unsquashable migration 20210101_01_abcd-first-migration [Y/n]: y

        """
        first migration
        """

        __depends__ = []
        __transaction__ = False

        async def apply(db):
            await db.execute("CREATE TABLE two();")

        async def rollback(db):
            await db.execute("DROP TABLE two;")

        Remove unsquashable migration 20210101_01_abcd-first-migration [y/N]: y
        '''),
        )

        assert sorted([path.stem for path in migrations.iterdir() if path.suffix in {".py", ".sql"}]) == []

    def test_keep_python_file(self, migration_file_factory, cli_runner, migrations):
        migration_file_factory(
            "20210101_01_abcd-first-migration",
            "py",
            dedent('''
            """
            first migration
            """
            __depends__ = []
            __transaction__ = False

            async def apply(db):
                await db.execute("CREATE TABLE two();")

            async def rollback(db):
                await db.execute("DROP TABLE two;")
            '''),
        )

        result = cli_runner.invoke(["squash", "--skip-prompt"], input="n\nn\n")
        assert result.exit_code == 0, result.output

        assert sorted([path.stem for path in migrations.iterdir() if path.suffix in {".py", ".sql"}]) == [
            "20210101_01_abcd-first-migration",
        ]

    def test_single_squash_skipped(self, migration_file_factory, cli_runner, migrations):
        migration_file_factory(
            "20210101_01_abcd-first-migration",
            "sql",
            dedent("""
            -- first migration
            -- depends:

            -- migrate: apply
            CREATE TABLE one (id INT);
            -- migrate: rollback
            DROP TABLE one;
            """),
        )

        result = cli_runner.invoke(["squash"])
        assert result.exit_code == 0, result.output

        assert sorted([path.stem for path in migrations.iterdir() if path.suffix in {".py", ".sql"}]) == [
            "20210101_01_abcd-first-migration",
        ]

    def test_backups_kept(self, migration_file_factory, cli_runner, migrations):
        migration_file_factory(
            "20210101_01_abcd-first-migration",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            CREATE TABLE one (id INT);
            -- migrate: rollback
            DROP TABLE one;
            """),
        )
        migration_file_factory(
            "20210101_02_efgh-second-migration",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_01_abcd-first-migration

            -- migrate: apply
            CREATE TABLE two (id INT);
            -- migrate: rollback
            DROP TABLE two;
            """),
        )

        result = cli_runner.invoke(["squash", "--backup"])
        assert result.exit_code == 0, result.output

        assert sorted([path.name for path in migrations.iterdir()]) == [
            "20210101_01_abcd-first-migration.sql.bak",
            "20210101_02_efgh-second-migration.sql",
            "20210101_02_efgh-second-migration.sql.bak",
        ]

    def test_sources_tracked(self, migration_file_factory, cli_runner):
        old = migration_file_factory(
            "20210101_01_abcd-first-migration",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            CREATE TABLE one (id INT);
            -- migrate: rollback
            DROP TABLE one;
            """),
        )
        new = migration_file_factory(
            "20210101_02_efgh-second-migration",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_01_abcd-first-migration

            -- migrate: apply
            CREATE TABLE two (id INT);
            -- migrate: rollback
            DROP TABLE two;
            """),
        )

        result = cli_runner.invoke(["squash", "--source"])
        assert result.exit_code == 0, result.output

        assert old.exists() is False
        assert new.read_text() == dedent("""\
        -- commit
        -- depends:

        -- squashed: 20210101_01_abcd-first-migration

        -- migrate: apply

        -- Squash one statements.

        CREATE TABLE one (id INT); -- source: 20210101_01_abcd-first-migration

        -- Squash two statements.

        CREATE TABLE two (id INT); -- source: 20210101_02_efgh-second-migration

        -- migrate: rollback

        -- Squash two statements.

        DROP TABLE two; -- source: 20210101_02_efgh-second-migration

        -- Squash one statements.

        DROP TABLE one; -- source: 20210101_01_abcd-first-migration
        """)



from unittest import mock

import pytest
from pogo_core.util import testing as core_testing

from pogo_migrate import testing
from tests.util import AsyncMock


@pytest.fixture(autouse=True)
def pyproject(pyproject_factory):
    return pyproject_factory()


async def test_apply(monkeypatch, db_session, cwd):
    monkeypatch.setattr(core_testing.migrate, "apply", AsyncMock())
    await testing.apply(db_session)

    assert core_testing.migrate.apply.call_args == mock.call(
        db_session,
        cwd / "migrations",
        schema_name="public",
    )


async def test_apply_loads_db(monkeypatch, cwd):
    monkeypatch.setattr(core_testing.migrate, "apply", AsyncMock())
    mock_session = mock.Mock(close=AsyncMock(return_value=None))
    monkeypatch.setattr(core_testing.sql, "get_connection", AsyncMock(return_value=mock_session))
    await testing.apply()

    assert core_testing.migrate.apply.call_args == mock.call(
        mock_session,
        cwd / "migrations",
        schema_name="public",
    )
    assert mock_session.close.call_count == 1


async def test_rollback(monkeypatch, db_session, cwd):
    monkeypatch.setattr(core_testing.migrate, "rollback", AsyncMock())
    await testing.rollback(db_session)

    assert core_testing.migrate.rollback.call_args == mock.call(
        db_session,
        cwd / "migrations",
        schema_name="public",
    )


async def test_rollback_loads_db(monkeypatch, cwd):
    monkeypatch.setattr(core_testing.migrate, "rollback", AsyncMock())
    mock_session = mock.Mock(close=AsyncMock(return_value=None))
    monkeypatch.setattr(core_testing.sql, "get_connection", AsyncMock(return_value=mock_session))
    await testing.rollback()

    assert core_testing.migrate.rollback.call_args == mock.call(
        mock_session,
        cwd / "migrations",
        schema_name="public",
    )
    assert mock_session.close.call_count == 1



from warnings import warn

from pogo_core.migration import (
    Migration,
    find_heads,
    read_sql_migration,
    strip_comments,
    terminate_statements,
    topological_sort,
)

warn(
    "pogo_migrate.migration usage has been deprecated, please use pogo_core.migration",
    FutureWarning,
    stacklevel=2,
)


__all__ = [
    "Migration",
    "find_heads",
    "read_sql_migration",
    "strip_comments",
    "terminate_statements",
    "topological_sort",
]
